from __future__ import annotations

from collections import defaultdict
from collections.abc import Callable, Iterator
from copy import copy, deepcopy
from itertools import zip_longest
from typing import Any, Generic

from lxml import etree

from ..array import argpass
from ..types import Accessor, EtreeFunction, T
from .attr import attr_constant, attr_function
from .bind import bind_index, bind_key
from .classed import class_array, classed_constant, classed_function
from .constant import constant
from .enter import EnterNode
from .matcher import matcher
from .namespace import namespace
from .style import style_constant, style_function, style_value
from .text import text_constant, text_function


def xpath(selection: str) -> str:
    """
    Returns a partial formatted xpath string given a selection

    Parameters
    ----------
    selection : str
        Selection string

    Returns
    -------
    str
        Partial formatted xpath
    """
    order = ""
    if ":" in selection:
        selection, order = selection.split(":")
        if order != "last-of-type":
            raise ValueError(
                f"Only 'last-of-type' is implemented currently (found {order})."
            )
        order = "[last()]"
    if "." in selection:
        tag, class_name = selection.split(".")
        tag = tag or "*"
        class_name = f"[@class='{class_name}']" if class_name else ""
        xpath = f"{tag}{order}{class_name}"
    elif "[" in selection and "]" in selection:
        tag, specifier = selection.split("[")
        tag = tag or "*"
        specifier = f"[@{specifier}" if specifier else ""
        xpath = f"{tag}{order}{specifier}"
    else:
        xpath = f"{selection}{order}"
    return xpath


def selector(
    element: etree.Element, selection: str | None = None
) -> list[etree.Element]:
    """
    Searchs recursively :code:`selection` where the root node is :code:`element`.

    Supported forms are:

    - :code:`{tag_name}{.class_name}{:last-of-type}`
    - :code:`{tag_name}{[attribute_name="value"]}{:last-of-type}`

    The :code:`tag_name` is any SVG element tag (ex: :code:`g`, :code:`rect`,
    :code:`line`, ...). You can chain selections by adding one space between
    them.

    The :code:`.class_name` is specified when a element is created such as:

    .. code:: python

        svg.append("g").attr("class", "my_class")

    In this example, the value of :code:`.class_name` is :code:`.my_class`.

    The :code:`:last-of-type` option is useful when you only need the last
    element.

    Using :code:`[attribute_name="value"]` allows you to get any element
    associated to a specific attribute. For instance, in the following
    code:

    .. code:: python

        svg.append("g").attr("aria-label", "my_group")

    To access this element, the value of :code:`[attribute_name="value"]`
    must be :code:`[aria-label="my_group"]`.

    Parameters
    ----------
    element : etree.Element
        Root node on which the search starts
    selection : str | None
        Selection description

    Returns
    -------
    list[etree.Element]
        List of found nodes.
    """
    if selection is None or selection == "":
        return element
    pxpath = (
        "/".join(map(xpath, selection.split(" ")))
        if " " in selection
        else xpath(selection)
    )
    return element.xpath(f"./*//{pxpath}") + element.xpath(f"./{pxpath}")


def creator(node: etree.Element, fullname: dict | None = None) -> etree.SubElement:
    """
    Creates a subnode associated to :code:`node`.

    Parameters
    ----------
    node : etree.Element
        Node on which the subnode will be associated.
    fullname : dict | None
        Dictionary with keys :code:`"local"` as name and :code:`"space"` for
        its namespace.

    Returns
    -------
    etree.SubElement
        Subnode
    """
    return (
        etree.SubElement(node, fullname["local"], nsmap={None: fullname["space"]})
        if isinstance(fullname, dict)
        else etree.SubElement(node, fullname, nsmap={})
    )


class Selection(Generic[T]):
    """
    A selection is a set of elements from the DOM. Typically these elements are
    identified by selectors such as .fancy for elements with the class fancy,
    or div to select DIV elements.

    Selection methods come in two forms, :code:`select` and :code:`select_all`:
    the former selects only the first matching element, while the latter
    selects all matching elements in document order. The top-level selection
    methods, d3.select and d3.select_all, query the entire document; the
    subselection methods, selection.select and selection.select_all, restrict
    selection to descendants of the selected elements.

    By convention, selection methods that return the current selection such as
    selection.attr use four spaces of indent, while methods that return a new
    selection use only two. This helps reveal changes of context by making them
    stick out of the chain:

    Parameters
    ----------
    groups : list[list[etree.Element]]
        List of groups of selected nodes given its parent.
    parents : list[etree.Element]
        List of parents related to groups.
    enter : list[EnterNode[T]] | None
        List of placeholder nodes for each datum that had no corresponding
        DOM element in the selection.
    exit : list[etree.Element] | None
        List of existing DOM elements in the selection for which no new datum was found.
    data : dict[etree.Element, T] | None
        Association between nodes and its data

    Examples
    --------

    >>> body = d3.create("body")
    >>> (
    ...     body
    ...     .append("svg")
    ...     .attr("width", 960)
    ...     .attr("height", 500)
    ...     .append("g")
    ...     .attr("transform", "translate(20, 20)")
    ...     .append("rect")
    ...     .attr("width", 920)
    ...     .attr("height", 460)
    ... )
    Selection(
        groups=[[rect]],
        parents=[g],
    )
    >>> print(body.to_string())
    <body>
      <svg xmlns="http://www.w3.org/2000/svg" weight="960" height="500">
        <g transform="translate(20, 20)">
          <rect width="920" height="460"/>
        </g>
      </svg>
    </body>
    >>> str(body)
    '<body><svg xmlns="http://www.w3.org/2000/svg" weight="960" height="500"><g transform="translate(20, 20)"><rect width="920" height="460"/></g></svg></body>'
    """

    def __init__(
        self,
        groups: list[list[etree.Element]],
        parents: list[etree.Element],
        enter: list[EnterNode[T]] | None = None,
        exit: list[etree.Element] | None = None,
        data: dict[etree.Element, T] | None = None,
    ):
        self._groups = groups
        self._parents = parents
        self._enter = enter
        self._exit = exit
        self._data = {} if data is None else data

    def select(self, selection: str | None = None) -> Selection:
        """
        Selects the first element that matches the specified :code:`selection` string.

        Supported forms are:

        - :code:`{tag_name}{.class_name}{:last-of-type}`
        - :code:`{tag_name}{[attribute_name="value"]}{:last-of-type}`

        The :code:`tag_name` is any SVG element tag (ex: :code:`g`,
        :code:`rect`, :code:`line`, ...). You can chain selections by adding
        one space between them.

        The :code:`.class_name` is specified when a element is created such as:

        .. code:: python

            svg.append("g").attr("class", "my_class")

        In this example, the value of :code:`.class_name` is :code:`.my_class`.

        The :code:`:last-of-type` option is useful when you only need the last
        element.

        Using :code:`[attribute_name="value"]` allows you to get any element
        associated to a specific attribute. For instance, in the following
        code:

        .. code:: python

            svg.append("g").attr("aria-label", "my_group")

        To access this element, the value of :code:`[attribute_name="value"]`
        must be :code:`[aria-label="my_group"]`.

        Parameters
        ----------
        selection : str | None
            Selection string

        Returns
        -------
        Selection
            Selection of first element

        Examples
        --------

        >>> svg = d3.create("svg")
        >>> svg.append("g").attr("class", "ticks")
        Selection(
            groups=[[g.ticks]],
            parents=[svg],
        )
        >>> svg.append("g").attr("class", "labels")
        Selection(
            groups=[[g.labels]],
            parents=[svg],
        )
        >>> svg.select("g.ticks")
        Selection(
            groups=[[g.ticks]],
            parents=[svg],
        )
        >>> svg.select("g.points")
        Selection(
            groups=[[]],
            parents=[svg],
        )
        """
        groups = defaultdict(list)
        nodes = (node for group in self._groups for node in group)
        for node in filter(lambda n: n is not None, nodes):
            if isinstance(node, EnterNode):
                node = node._parent
            subgroup = selector(node, selection)[:1]
            if len(subgroup) == 0:
                groups[node]
                continue
            for subnode in subgroup:
                parent = subnode.getparent()
                groups[parent].append(subnode)

        subgroups = list(groups.values())
        parents = list(groups)

        return Selection(subgroups, parents, data=self._data)

    def select_all(self, selection: str | None = None) -> Selection:
        """
        Selects all elements that match the specified :code:`selection` string.

        Supported forms are:

        - :code:`{tag_name}{.class_name}{:last-of-type}`
        - :code:`{tag_name}{[attribute_name="value"]}{:last-of-type}`

        The :code:`tag_name` is any SVG element tag (ex: :code:`g`,
        :code:`rect`, :code:`line`, ...). You can chain selections by adding
        one space between them.

        The :code:`.class_name` is specified when a element is created such as:

        .. code:: python

            svg.append("g").attr("class", "my_class")

        In this example, the value of :code:`.class_name` is :code:`.my_class`.

        The :code:`:last-of-type` option is useful when you only need the last
        element.

        Using :code:`[attribute_name="value"]` allows you to get any element
        associated to a specific attribute. For instance, in the following
        code:

        .. code:: python

            svg.append("g").attr("aria-label", "my_group")

        To access this element, the value of :code:`[attribute_name="value"]`
        must be :code:`[aria-label="my_group"]`.

        Parameters
        ----------
        selection : str | None
            Selection string

        Returns
        -------
        Selection
            Selection of all matched elements

        Examples
        --------

        >>> svg = d3.create("svg")
        >>> scale = d3.scale_linear([0, 10], [0, 100])
        >>> print(svg.call(d3.axis_bottom(scale).set_ticks(3)).to_string())
        <svg xmlns:xmlns="http://www.w3.org/2000/svg" fill="none" font-size="10" font-family="sans-serif" text-anchor="middle">
          <path class="domain" stroke="currentColor" d="M0.5,6V0.5H100.5V6"/>
          <g class="tick" opacity="1" transform="translate(0.5, 0)">
            <line stroke="currentColor" y2="6"/>
            <text fill="currentColor" y="9" dy="0.71em">0</text>
          </g>
          <g class="tick" opacity="1" transform="translate(50.5, 0)">
            <line stroke="currentColor" y2="6"/>
            <text fill="currentColor" y="9" dy="0.71em">5</text>
          </g>
          <g class="tick" opacity="1" transform="translate(100.5, 0)">
            <line stroke="currentColor" y2="6"/>
            <text fill="currentColor" y="9" dy="0.71em">10</text>
          </g>
        </svg>
        >>> svg.select_all("g").select_all("line")
        Selection(
            groups=[[line], [line], [line]],
            parents=[g.tick, g.tick, g.tick],
        )
        >>> svg.select_all("line")
        Selection(
            groups=[[line], [line], [line]],
            parents=[g.tick, g.tick, g.tick],
        )
        """
        groups = defaultdict(list)
        nodes = (node for group in self._groups for node in group)
        for node in filter(lambda n: n is not None, nodes):
            if isinstance(node, EnterNode):
                node = node._parent
            subgroup = selector(node, selection)
            if len(subgroup) == 0:
                groups[node]
                continue
            for subnode in subgroup:
                parent = subnode.getparent()
                groups[parent].append(subnode)

        subgroups = list(groups.values())
        parents = list(groups)

        return Selection(subgroups, parents, data=self._data)

    def enter(self) -> Selection:
        """
        Returns the enter selection: placeholder nodes for each datum that had
        no corresponding DOM element in the selection.

        Returns
        -------
        Selection
            Enter selection

        Examples
        --------

        >>> svg = d3.create("svg")
        >>> text = svg.select_all("text").data(["hello", "world"])
        >>> text_enter = text.enter()
        >>> text_enter
        Selection(
            groups=[[EnterNode(svg, hello), EnterNode(svg, world)]],
            parents=[svg],
        )
        """
        return Selection(
            self._enter or [[None] * len(group) for group in self._groups],
            self._parents,
            data=self._data,
        )

    def exit(self) -> Selection:
        """
        Returns the exit selection: existing DOM elements in the selection for
        which no new datum was found. (The exit selection is empty for
        selections not returned by selection.data.)

        Returns
        -------
        Selection
            Exit selection

        Examples
        --------

        >>> svg = d3.create("svg")
        >>> text = svg.select_all("text").data(["hello", "world"])
        >>> text_exit = text.exit()
        >>> text_exit
        Selection(
            groups=[[]],
            parents=[svg],
        )
        """
        return Selection(
            self._exit or [[None] * len(group) for group in self._groups],
            self._parents,
            data=self._data,
        )

    def merge(self, context: Selection) -> Selection:
        """
        Returns a new selection merging this selection with the specified other
        selection or transition. The returned selection has the same number of
        groups and the same parents as this selection. Any missing (None)
        elements in this selection are filled with the corresponding element,
        if present (not null), from the specified selection. (If the other
        selection has additional groups or parents, they are ignored.)

        Parameters
        ----------
        context : Selection
            Selection

        Returns
        -------
        Selection
            Merged selection

        Examples
        --------

        >>> svg = d3.create("svg")
        >>> text = svg.select_all("text").data(["hello", "world"])
        >>> text_enter = text.enter()
        >>> text_enter.append("text").text(lambda text: text)
        Selection(
            groups=[[text, text]],
            parents=[svg],
        )
        >>> print(svg.to_string())
        <svg xmlns:xmlns="http://www.w3.org/2000/svg">
          <text>hello</text>
          <text>world</text>
        </svg>
        >>> print(text.to_repr())
        Selection(
            groups=[[None, None]],
            parents=[svg],
            enter=[[EnterNode(svg, hello), EnterNode(svg, world)]],
            exit=[[]],
            data={<Element text at 0x7d28b3f08380>: 'hello', <Element text at 0x7d28b3f08140>: 'world'},
        )
        >>> text = text.merge(text_enter)
        >>> print(text.to_repr())
        Selection(
            groups=[[EnterNode(svg, hello), EnterNode(svg, world)]],
            parents=[svg],
            enter=None,
            exit=None,
            data={<Element text at 0x7d28b3f08380>: 'hello', <Element text at 0x7d28b3f08140>: 'world'},
        )
        """
        selection = context.selection() if hasattr(context, "selection") else context

        merges = []
        for groups0, groups1 in zip_longest(
            self._groups, selection._groups, fillvalue=[]
        ):
            merge = []
            for group0, group1 in zip_longest(groups0, groups1, fillvalue=None):
                node = group0 if group0 is not None else group1
                merge.append(node)
            merges.append(merge)

        for j in range(len(merges), len(self._groups)):
            merges.append(self._groups[j])

        return Selection(merges, self._parents, data=self._data | selection._data)

    def filter(self, match: Accessor[T, bool] | int | float | str) -> Selection:
        """
        Filters the selection, returning a new selection that contains only the
        elements for which the specified filter is true.

        Parameters
        ----------
        match : Accessor[T, bool] | int | float | str
            Constant to match or accessor which returns a boolean

        Returns
        -------
        Selection
            Filtered selection

        Examples
        --------

        >>> svg = d3.create("svg")
        >>> scale = d3.scale_linear([0, 10], [0, 100])
        >>> print(svg.call(d3.axis_bottom(scale).set_ticks(3)).to_string())
        <svg xmlns="http://www.w3.org/2000/svg" fill="none" font-size="10" font-family="sans-serif" text-anchor="middle">
          <path class="domain" stroke="currentColor" d="M0.5,6V0.5H100.5V6"/>
          <g class="tick" opacity="1" transform="translate(0.5, 0)">
            <line stroke="currentColor" y2="6"/>
            <text fill="currentColor" y="9" dy="0.71em">0</text>
          </g>
          <g class="tick" opacity="1" transform="translate(50.5, 0)">
            <line stroke="currentColor" y2="6"/>
            <text fill="currentColor" y="9" dy="0.71em">5</text>
          </g>
          <g class="tick" opacity="1" transform="translate(100.5, 0)">
            <line stroke="currentColor" y2="6"/>
            <text fill="currentColor" y="9" dy="0.71em">10</text>
          </g>
        </svg>
        >>> result = svg.select_all("text").filter(lambda d, i: i % 2 != 0)
        >>> result
        Selection(
            groups=[[text]],
            parents=[svg],
        )
        >>> result.node().text
        '5'
        """
        matches = matcher(match)
        subgroups = []
        for group in self._groups:
            subgroup = []
            for i, node in enumerate(group):
                if node is None:
                    continue
                if isinstance(node, EnterNode):
                    node = node._parent
                if matches(self._data.get(node), i, group):
                    subgroup.append(node)
            subgroups.append(subgroup)
        return Selection(subgroups, self._parents, data=self._data)

    def append(self, name: str) -> Selection:
        """
        If the specified name is a string, appends a new element of this type
        (tag name) as the last child of each selected element, or before the
        next following sibling in the update selection if this is an enter
        selection. The latter behavior for enter selections allows you to
        insert elements into the DOM in an order consistent with the new bound
        data; however, note that selection.order may still be required if
        updating elements change order (i.e., if the order of new data is
        inconsistent with old data).

        Parameters
        ----------
        name : str
            Tag name

        Returns
        -------
        Selection
            Selection

        Examples
        --------

        Simple append:

        >>> svg = d3.create("svg")
        >>> print(svg.to_string())
        <svg xmlns="http://www.w3.org/2000/svg"/>
        >>> g = svg.append("g").attr("class", "labels")
        >>> print(svg.to_string())
        <svg xmlns="http://www.w3.org/2000/svg">
          <g class="labels"/>
        </svg>

        Multiple append:

        >>> import detroit as d3
        >>> svg = d3.create("svg")
        >>> (
        ...     svg.select_all()
        ...     .data([None] * 2)
        ...     .enter()
        ...     .append("g")
        ...     .append("text")
        ... )
        Selection(
            groups=[[text], [text]],
            parents=[g, g],
        )
        >>> print(svg.to_string())
        <svg xmlns="http://www.w3.org/2000/svg">
          <g>
            <text/>
          </g>
          <g>
            <text/>
          </g>
        </svg>
        """
        fullname = namespace(name)
        groups = defaultdict(list)
        nodes = (node for group in self._groups for node in group)
        for node in filter(lambda n: n is not None, nodes):
            if isinstance(node, EnterNode):
                enter_node = node
                node = enter_node._parent
                subnode = creator(node, fullname)
                node.append(subnode)
                self._data[subnode] = enter_node.__data__
                groups[node].append(subnode)
            else:
                subnode = creator(node, fullname)
                node.append(subnode)
                groups[node].append(subnode)
                self._data[subnode] = self._data.get(node)
        subgroups = list(groups.values())
        parents = list(groups)
        return Selection(subgroups, parents, data=self._data)

    def each(self, callback: EtreeFunction[T, None]) -> Selection:
        """
        Invokes the specified function for each selected element, in order,
        being passed the current DOM element (nodes[i]), the current datum (d),
        the current index (i), and the current group (nodes).

        Parameters
        ----------
        callback : EtreeFunction[T, None]
            Function to call which takes as argument:

            * **node** (:code:`etree.Element`) - the node element
            * **data** (:code:`T`) - current data associated to the node
            * **index** (:code:`int`) - the index of the node in its group
            * **group** (:code:`list[etree.Element]`) - the node's group with other nodes.

        Returns
        -------
        Selection
            Itself

        Examples
        --------

        >>> import detroit as d3
        >>> svg = d3.create("svg")
        >>> (
        ...     svg.select_all()
        ...     .data(list(range(4)))
        ...     .enter()
        ...     .append("g")
        ...     .attr("transform", lambda d: f"scale({d})")
        ... )
        Selection(
            groups=[[g, g, g, g]],
            parents=[svg],
        )
        >>> def each(node, d, i, group):
        ...     tag = node.tag
        ...     transform = node.attrib.get("transform")
        ...     print(f"{tag = }, {transform = }, data = {d}, {i = }, {len(group) = }")
        ...
        >>> svg.select_all("g").each(each)
        tag = 'g', transform = 'scale(0)', data = 0, i = 0, len(group) = 4
        tag = 'g', transform = 'scale(1)', data = 1, i = 1, len(group) = 4
        tag = 'g', transform = 'scale(2)', data = 2, i = 2, len(group) = 4
        tag = 'g', transform = 'scale(3)', data = 3, i = 3, len(group) = 4
        Selection(
            groups=[[g, g, g, g]],
            parents=[svg],
        )
        """
        callback = argpass(callback)
        for group in self._groups:
            for i, node in enumerate(group):
                if node is not None:
                    if isinstance(node, EnterNode):
                        node = node._parent
                    callback(node, self._data.get(node), i, group)
        return self

    def attr(
        self, name: str, value: Accessor[T, Any] | list[Any] | Any | None = None
    ) -> Selection:
        """
        If a value is specified, sets the attribute with the specified name to
        the specified value on the selected elements and returns this
        selection.

        If the value is a function, it is evaluated for each selected element,
        in order, being passed the current datum (d), the current index (i),
        and the current group (nodes).

        Parameters
        ----------
        name : str
            Name of the attribute
        value : Accessor[T, Any] | list[Any] | Any | None
            Value function or constant value. The final value is converted to a
            string.

        Returns
        -------
        Selection
            Itself

        Examples
        --------

        >>> import detroit as d3
        >>> svg = d3.create("svg")
        >>> print(
        ...     svg.append("g")
        ...     .attr("class", "labels")
        ...     .attr("transform", "translate(20, 10)")
        ...     .to_string()
        ... )
        <svg xmlns="http://www.w3.org/2000/svg">
          <g class="labels" transform="translate(20, 10)"/>
        </svg>
        """
        fullname = namespace(name)
        name = (
            f"{{{fullname['space']}}}{fullname['local']}"
            if isinstance(fullname, dict)
            else fullname
        )
        if value is None:
            return self.node().get(name)
        elif callable(value):
            self.each(attr_function(name, value))
        else:
            self.each(attr_constant(name, value))
        return self

    def property(
        self, name: str, value: Accessor[T, Any] | list[Any] | Any | None = None
    ) -> Selection:
        """
        This method has no difference with :code:`Selection.attr`.

        Parameters
        ----------
        name : str
            Name of the property
        value : Accessor[T, Any] | list[Any] | Any | None
            Property value function or constant property value. The final value
            is converted to a string.

        Returns
        -------
        Selection
            Itself

        Examples
        --------

        >>> import detroit as d3
        >>> svg = d3.create("svg")
        >>> print(
        ...     svg.append("g")
        ...     .property("class", "labels")
        ...     .property("transform", "translate(20, 10)")
        ...     .to_string()
        ... )
        <svg xmlns="http://www.w3.org/2000/svg">
          <g class="labels" transform="translate(20, 10)"/>
        </svg>
        """
        return self.attr(name, value)

    def style(
        self, name: str, value: Accessor[T, str] | str | None = None
    ) -> Selection:
        """
        If a value is specified, sets the style with the specified name to the
        specified value on the selected elements and returns this selection.

        If the value is a function, it is evaluated for each selected element,
        in order, being passed the current datum (d), the current index (i),
        and the current group (nodes).

        Parameters
        ----------
        name : str
            Name of the style
        value : Accessor[T, str] | str | None
            Value function or constant value.

        Returns
        -------
        Selection
            Itself

        Examples
        --------

        >>> import detroit as d3
        >>> svg = d3.create("svg")
        >>> print(
        ...     svg.append("text")
        ...     .style("fill", "black")
        ...     .style("stroke", "none")
        ...     .to_string()
        ... )
        <svg xmlns="http://www.w3.org/2000/svg">
          <text style="fill:black;stroke:none;"/>
        </svg>
        """
        if value is None:
            return style_value(self.node().get("style"), name)
        elif callable(value):
            self.each(style_function(name, value))
        else:
            self.each(style_constant(name, value))
        return self

    def text(self, value: Accessor[T, Any] | Any | None = None) -> Selection:
        """
        If the value is a constant, then all elements are given the same text
        content; otherwise, if the value is a function, it is evaluated for
        each selected element, in order, being passed the current datum (d),
        the current index (i), and the current group (nodes).

        Parameters
        ----------
        value : Accessor[T, Any] | Any | None
            Text function or constant text value. The final value is converted
            to a string.

        Returns
        -------
        Selection
            Itself

        Examples
        --------

        Direct assignment:

        >>> import detroit as d3
        >>> svg = d3.create("svg")
        >>> print(svg.append("text").text("Hello, world!").to_string())
        <svg xmlns="http://www.w3.org/2000/svg">
          <text>Hello, world!</text>
        </svg>

        Through data:

        >>> import detroit as d3
        >>> svg = d3.create("svg")
        >>> print(
        ...     svg.select_all("text")
        ...     .data(["Hello", "world"])
        ...     .enter()
        ...     .append("text")
        ...     .text(lambda text, i: f"{text} - index {i}")
        ...     .to_string()
        ... )
        <svg xmlns="http://www.w3.org/2000/svg">
          <text>Hello - index 0</text>
          <text>world - index 1</text>
        </svg>
        """
        if value is None:
            return self.node().text
        elif callable(value):
            self.each(text_function(value))
        else:
            self.each(text_constant(value))
        return self

    def html(self, value: Accessor[T, Any] | Any | None = None) -> Selection:
        """
        This method has no difference with :code:`Selection.text`.

        Parameters
        ----------
        value : Accessor[T, Any] | Any | None
            Inner HML function or constant inner HML. The final value is
            converted to a string.

        Returns
        -------
        Selection
            Itself

        Examples
        --------

        Direct assignment:

        >>> import detroit as d3
        >>> svg = d3.create("svg")
        >>> print(svg.append("text").html("Hello, world!").to_string())
        <svg xmlns="http://www.w3.org/2000/svg">
          <text>Hello, world!</text>
        </svg>

        Through data:

        >>> import detroit as d3
        >>> svg = d3.create("svg")
        >>> print(
        ...     svg.select_all("text")
        ...     .data(["Hello", "world"])
        ...     .enter()
        ...     .append("text")
        ...     .html(lambda text, i: f"{text} - index {i}")
        ...     .to_string()
        ... )
        <svg xmlns="http://www.w3.org/2000/svg">
          <text>Hello - index 0</text>
          <text>world - index 1</text>
        </svg>
        """
        return self.text(value)

    def classed(
        self, names: str, value: Accessor[T, bool] | bool | None = None
    ) -> Selection:
        """
        Assigns or unassigns the specified CSS class names on the selected
        elements by setting the class attribute or modifying the class list
        property and returns this selection.

        Parameters
        ----------
        names : str
            Class names
        value : Accessor[T, bool] | bool | None
            Boolean function or constant boolean where the boolean indicates if
            the class names must be added or removed from the class property of
            the node.

        Returns
        -------
        Selection
            Itself

        Examples
        --------

        >>> import detroit as d3
        >>> svg = d3.create("svg")
        >>> data = ["Hello", "world"]
        >>> (
        ...    svg.select_all()
        ...    .data(data)
        ...    .enter()
        ...    .append("g")
        ...    .classed("myclass", lambda d: d == "Hello")
        ... )
        Selection(
            groups=[[g.myclass, g]],
            parents=[svg],
        )
        >>> str(svg)
        '<svg xmlns="http://www.w3.org/2000/svg"><g class="myclass"/><g/></svg>'
        """
        if value is None:
            node = self.node()
            if class_string := node.get("class"):
                names = class_array(names)
                classes = class_array(class_string)
                return len(set(names) & set(classes)) == len(names)
            return False
        elif callable(value):
            self.each(classed_function(names, value))
        else:
            self.each(classed_constant(names, value))
        return self

    def datum(self, value: T) -> Selection:
        """
        Sets the bound data for the first selected node.

        Parameters
        ----------
        value : T
            Value

        Returns
        -------
        Selection
            Itself

        Examples
        --------

        >>> import detroit as d3
        >>> svg = d3.create("svg")
        >>> g1 = svg.append("g").attr("class", "g1")
        >>> g2 = svg.append("g").attr("class", "g2")
        >>> g = svg.select_all("g")
        >>> print(g.to_repr())
        Selection(
            groups=[[g.g1, g.g2]],
            parents=[svg],
            enter=None,
            exit=None,
            data={<Element g at 0x7d115a460a80>: None, <Element g at 0x7d115a460640>: None},
        )
        >>> g.datum("Hello, world")
        Selection(
            groups=[[g.g1, g.g2]],
            parents=[svg],
        )
        >>> print(g.to_repr())
        Selection(
            groups=[[g.g1, g.g2]],
            parents=[svg],
            enter=None,
            exit=None,
            data={<Element g at 0x7d115a460a80>: 'Hello, world', <Element g at 0x7d115a460640>: None},
        )
        >>> g1.node()
        <Element g at 0x7d115a460a80>
        """
        self._data[self.node()] = value
        return self

    def data(
        self,
        values: list[T] | EtreeFunction[T, list[T]],
        key: Accessor[T, float | str] | None = None,
    ) -> Selection:
        """
        Binds the specified list of data with the selected elements, returning
        a new selection that represents the update selection: the elements
        successfully bound to data. Also defines the enter and exit selections
        on the returned selection, which can be used to add or remove elements
        to correspond to the new data. The specified data is an array of
        arbitrary values (e.g., numbers or objects), or a function that returns
        an array of values for each group.

        Parameters
        ----------
        values : list[T] | EtreeFunction[T, list[T]]
            List of data to bind
        key : Accessor[T, float | str] | None
            Optional accessor which returns a key value

        Returns
        -------
        Selection
            Itself

        Examples
        --------

        >>> import detroit as d3
        >>> svg = d3.create("svg")
        >>> print(svg.append("g").to_repr())
        Selection(
            groups=[[g]],
            parents=[svg],
            enter=None,
            exit=None,
            data={<Element g at 0x75cf61141b80>: None},
        )
        >>> print(svg.append("g").to_repr())
        Selection(
            groups=[[g]],
            parents=[svg],
            enter=None,
            exit=None,
            data={<Element g at 0x75cf61141b80>: None, <Element g at 0x75cf611b0500>: None},
        )
        >>> print(svg.select_all("text").data(["Hello", "world"]).to_repr())
        Selection(
            groups=[[None, None]],
            parents=[svg],
            enter=[[EnterNode(svg, Hello), EnterNode(svg, world)]],
            exit=[[]],
            data={<Element g at 0x75cf61141b80>: None, <Element g at 0x75cf611b0500>: None},
        )
        >>> print(svg.select_all("g").data(["Hello", "world"]).to_repr())
        Selection(
            groups=[[g, g]],
            parents=[svg],
            enter=[[None, None]],
            exit=[[None, None]],
            data={<Element g at 0x75cf61141b80>: 'Hello', <Element g at 0x75cf611b0500>: 'world'},
        )
        """
        bind = bind_key if key else bind_index
        parents = self._parents
        groups = self._groups

        if not callable(values):
            values = constant(values)
        values = argpass(values)

        update = [None] * len(groups)
        enter = [None] * len(groups)
        exit = [None] * len(groups)
        for j in range(len(groups)):
            parent = parents[j]
            group = groups[j]
            data = list(values(parent, self._data.get(parent), j, parents))
            enter[j] = enter_group = [None] * len(data)
            update[j] = update_group = [None] * len(data)
            exit[j] = exit_group = [None] * len(group)

            bind(
                self._data,
                parent,
                group,
                enter_group,
                update_group,
                exit_group,
                data,
                key,
            )

            i1 = 0
            for i0 in range(len(data)):
                if previous := enter_group[i0]:
                    if i0 >= i1:
                        i1 = i0 + 1
                    while not (
                        i1 < len(update_group) and update_group[i1] is not None
                    ) and i1 < len(data):
                        i1 += 1
                    previous._next = update_group[i1] if i1 < len(data) else None

        return Selection(update, parents, enter, exit, self._data)

    def order(self) -> Selection:
        """
        Re-inserts elements into the document such that the document order of
        each group matches the selection order.

        Returns
        -------
        Selection
            Itself
        """
        for group in self._groups:
            next_node = None
            for node in reversed(group):
                if node is not None:
                    if next_node is not None and (node.getprevious() == next_node):
                        parent = next_node.getparent()
                        index = parent.index(next_node)
                        parent.pop(index)
                        parent.insert(index, node)
                    next_node = node
        return self

    def join(
        self,
        onenter: Callable[[Selection], Selection] | str,
        onupdate: Callable[[Selection], Selection] | None = None,
        onexit: Callable[[Selection], None] | None = None,
    ) -> Selection:
        """
        Appends, removes and reorders elements as necessary to match the data
        that was previously bound by selection.data, returning the merged enter
        and update selection. This method is a convenient alternative to the
        explicit general update pattern, replacing selection.enter,
        selection.exit, selection.append, selection.remove, and
        selection.order.

        Parameters
        ----------
        onenter : Callable[[Selection], Selection] | str
            Enter selection or function
        onupdate : Callable[[Selection], Selection] | None
            Function
        onexit : Callable[[Selection], None] | None
            Function

        Returns
        -------
        Selection
            Selection with joined elements

        Examples
        --------

        >>> import detroit as d3
        >>> data = [[0, 1, 2, 3], [4, 5, 6, 7], [8, 9, 10, 11], [12, 13, 14, 15]]
        >>> svg = d3.create("svg")
        >>> table = (
        ...     svg.append("table")
        ...     .select_all("tr")
        ...     .data(data)
        ...     .join("tr")
        ...     .select_all("td")
        ...     .data(lambda _, d: d)
        ...     .join("td")
        ...     .text(lambda d: str(d))
        ... )
        >>> print(svg.to_string())
        <svg xmlns="http://www.w3.org/2000/svg">
          <table>
            <tr>
              <td>0</td>
              <td>1</td>
              <td>2</td>
              <td>3</td>
            </tr>
            <tr>
              <td>4</td>
              <td>5</td>
              <td>6</td>
              <td>7</td>
            </tr>
            <tr>
              <td>8</td>
              <td>9</td>
              <td>10</td>
              <td>11</td>
            </tr>
            <tr>
              <td>12</td>
              <td>13</td>
              <td>14</td>
              <td>15</td>
            </tr>
          </table>
        </svg>

        Another usage could be to specify functions :

        >>> data = [None] * 3
        >>> svg = d3.create("svg")
        >>> svg.append("circle").attr("fill", "yellow")
        Selection(
            groups=[[circle]],
            parents=[svg],
        )
        >>> print(svg.to_string())
        <svg xmlns="http://www.w3.org/2000/svg">
          <circle fill="yellow"/>
        </svg>
        >>> (
        ...     svg.select_all("circle")
        ...     .data(data)
        ...     .join(
        ...         onenter=lambda enter: enter.append("circle").attr("fill", "green"),
        ...         onupdate=lambda update: update.attr("fill", "blue")
        ...     )
        ...     .attr("stroke", "black")
        ... )
        Selection(
            groups=[[circle, circle, None]],
            parents=[svg],
        )
        >>> print(svg.to_string())
        <svg xmlns="http://www.w3.org/2000/svg">
          <circle fill="blue"/>
          <circle fill="green" stroke="black"/>
          <circle fill="green" stroke="black"/>
        </svg>

        In this example, the attribute :code:`fill` of the existing circle was
        updated, by :code:`onupdate`, from :code:`yellow` to :code:`blue`. And
        since :code:`data` has 3 elements, :code:`onenter` has generated the
        last circles.
        """
        enter = self.enter()
        update = self
        exit = self.exit()

        # Enter
        if callable(onenter):
            enter = onenter(enter)
            if enter:
                enter = enter.selection()
        else:
            enter = enter.append(onenter)

        # Update
        if onupdate is not None:
            update = onupdate(update)
            if update:
                update = update.selection()

        # Exit
        if onexit is None:
            exit.remove()
        else:
            onexit(exit)

        if enter and update:
            return enter.merge(update).order()
        else:
            return update

    def insert(self, name: str, before: str) -> Selection:
        """
        If the specified name is a string, inserts a new element of this type
        (tag name) before the first element matching the specified
        :code:`before` selector for each selected element.

        Parameters
        ----------
        name : str
            Tag name
        before : str
            Node element selection

        Returns
        -------
        Selection
            Selection with inserted element(s)

        Examples
        --------

        Since :code:`before` refers to a selected tag, this method will operate
        multiple insertions.

        >>> import detroit as d3
        >>> svg = d3.create("svg")
        >>> (
        ...     svg.select_all("g")
        ...     .data([None, None])
        ...     .enter()
        ...     .append("g")
        ...     .append("text")
        ... )
        Selection(
            groups=[[text], [text]],
            parents=[g, g],
        )
        >>> print(svg.to_string())
        <svg xmlns="http://www.w3.org/2000/svg">
          <g>
            <text/>
          </g>
          <g>
            <text/>
          </g>
        </svg>
        >>> svg.insert("circle", "text")
        Selection(
            groups=[[circle], [circle]],
            parents=[g, g],
        )
        >>> print(svg.to_string())
        <svg xmlns="http://www.w3.org/2000/svg">
          <g>
            <circle/>
            <text/>
          </g>
          <g>
            <circle/>
            <text/>
          </g>
        </svg>

        If you prefer to insert an element at a specific location, you need to
        select first the specific node and then insert your element.

        >>> svg = d3.create("svg")
        >>> (
        ...     svg.select_all("g")
        ...     .data(["class1", "class2"])
        ...     .enter()
        ...     .append("g")
        ...     .append("text")
        ...     .attr("class", lambda d: d)
        ... )
        Selection(
            groups=[[text.class1], [text.class2]],
            parents=[g, g],
        )
        >>> svg.insert("circle", ".class1")
        Selection(
            groups=[[circle]],
            parents=[g],
            enter=None,
            exit=None,
            data={<Element circle at 0x7fc5753f0500>: None},
        )
        >>> print(svg.to_string())
        <svg xmlns="http://www.w3.org/2000/svg">
          <g>
            <circle/>
            <text class="class1"/>
          </g>
          <g>
            <text class="class2"/>
          </g>
        </svg>
        """
        fullname = namespace(name)
        selection = self.select_all(before)
        subgroups = []
        for i, group in enumerate(selection._groups):
            if len(group) > 0:
                node = group[0]
                parent = selection._parents[i]
                index = parent.index(node)
                created = creator(parent, fullname)
                parent.insert(index, created)
                self._data[created] = self._data.get(node)
                subgroup = [created]
            else:
                subgroup = []
            subgroups.append(subgroup)

        return Selection(subgroups, selection._parents, data=self._data)

    def remove(self) -> Selection:
        """
        Removes the selected elements from the document. Returns this selection
        (the removed elements) which are now detached from the DOM.

        Returns
        -------
        Selection
            Selection with removed elements

        Examples
        --------

        >>> import detroit as d3
        >>> svg = d3.create("svg")
        >>> (
        ...     svg.select_all("g")
        ...     .data([None] * 10)
        ...     .enter()
        ...     .append("g")
        ...     .attr("class", "domain")
        ... )
        Selection(
            groups=[[g.domain, g.domain, g.domain, g.domain, g.domain, g.domain, g.domain, g.domain, g.domain, g.domain]],
            parents=[svg],
        )
        >>> print(svg.to_string())
        <svg xmlns="http://www.w3.org/2000/svg">
          <g class="domain"/>
          <g class="domain"/>
          <g class="domain"/>
          <g class="domain"/>
          <g class="domain"/>
          <g class="domain"/>
          <g class="domain"/>
          <g class="domain"/>
          <g class="domain"/>
          <g class="domain"/>
        </svg>
        >>> svg.select_all(".domain").remove()
        Selection(
            groups=[[]],
            parents=[svg],
            enter=None,
            exit=None,
            data={},
        )
        >>> print(svg.to_string())
        <svg xmlns="http://www.w3.org/2000/svg"/>
        """
        subgroups = []
        for group in self._groups:
            subgroup = []
            for node in group:
                if node is not None:
                    if isinstance(node, EnterNode):
                        node = node._parent
                    subgroup.append(node)
                    parent = node.getparent()
                    if parent is not None:
                        parent.remove(node)
                        subgroup.pop()
                        self._data.pop(node, None)
            subgroups.append(subgroup)
        return Selection(subgroups, self._parents, data=self._data)

    def call(self, func: Callable[[Selection, ...], Any], *args: Any) -> Selection:
        """
        Invokes the specified function exactly once, passing in this selection
        along with any optional arguments. Returns this selection.

        Parameters
        ----------
        func : Callable[[Selection, ...], Any]
            Function to call
        args : Any
            Arguments for the function to call

        Returns
        -------
        Selection
            Itself

        Examples
        --------

        This is equivalent to invoking the function by hand but facilitates
        method chaining. For example, to set several styles in a reusable
        function:

        >>> def name(selection, first, last):
        ...     selection.attr("first-name", first).attr("last-name", last)

        Now say:

        >>> d3.select_all("div").call(name, "John", "Snow")

        This is roughly equivalent to:

        >>> name(d3.select_all("div"), "John", "Snow")
        """
        func(self, *args)
        return self

    def clone(self, deep: bool = False) -> Selection:
        """
        Returns a clone of the selection.

        Parameters
        ----------
        deep : bool
            :code:`True` for deep copy

        Returns
        -------
        Selection
            Clone of itself

        Notes
        -----
        Data, enter nodes, exit nodes are also cloned.

        Examples
        --------
        >>> import detroit as d3
        >>> svg = d3.create("svg")
        >>> clone = svg.clone()
        >>> rect = svg.append("rect")
        >>> clone.select_all("rect")
        Selection(
            groups=[[]],
            parents=[svg],
        )
        >>> g = clone.append("g")
        >>> svg.select_all("rect")
        Selection(
            groups=[[rect]],
            parents=[svg],
        )
        >>> svg.select_all("g")
        Selection(
            groups=[[]],
            parents=[svg],
        )
        """
        copy_func = deepcopy if deep else copy

        # Clone data and groups
        cloned_data = {}
        cloned_groups = []
        for group in self._groups:
            cloned_group = []
            for node in group:
                if isinstance(node, EnterNode):
                    cloned_group.append(node.clone(deep))
                    continue
                cloned_node = copy_func(node)
                if node in self._data:
                    cloned_data[cloned_node] = copy_func(self._data[node])
                cloned_group.append(cloned_node)
            cloned_groups.append(cloned_group)

        # Clone parents
        cloned_parents = [copy_func(parent) for parent in self._parents]
        cloned_enter = (
            None if self._enter is None else [node.clone(deep) for node in self._enter]
        )
        cloned_exit = (
            None if self._exit is None else [copy_func(node) for node in self._exit]
        )

        return Selection(
            cloned_groups,
            cloned_parents,
            enter=cloned_enter,
            exit=cloned_exit,
            data=cloned_data,
        )

    def node(self) -> etree.Element:
        """
        Returns the first (non-None) element in this selection.

        Returns
        -------
        etree.Element
            Node

        Examples
        --------

        >>> import detroit as d3
        >>> svg = d3.create("svg")
        >>> g = (
        ...     svg.select_all("g")
        ...     .data(list(reversed(range(10))))
        ...     .enter()
        ...     .append("g")
        ...     .attr("class", lambda d: f"class{d}")
        ... )
        >>> g
        Selection(
            groups=[[g.class9, g.class8, g.class7, g.class6, g.class5, g.class4, g.class3, g.class2, g.class1, g.class0]],
            parents=[svg],
        )
        >>> first = g.node()
        >>> print(f"{first.tag}.{first.attrib.get("class")}")
        g.class9
        """
        return next(iter(self))

    def nodes(self) -> list[etree.Element]:
        """
        Returns a list of all (non-None) elements in this selection.

        Returns
        -------
        list[etree.Element]
            List of nodes

        Examples
        --------
        >>> svg = d3.create("svg")
        >>> g = (
        ...     svg.select_all("g")
        ...     .data(list(reversed(range(10))))
        ...     .enter()
        ...     .append("g")
        ...     .attr("class", lambda d: f"class{d}")
        ... )
        >>> g
        Selection(
            groups=[[g.class9, g.class8, g.class7, g.class6, g.class5, g.class4, g.class3, g.class2, g.class1, g.class0]],
            parents=[svg],
        )
        >>> [f"{node.tag}.{node.attrib.get("class")}" for node in g.nodes()]
        ['g.class9', 'g.class8', 'g.class7', 'g.class6', 'g.class5', 'g.class4', 'g.class3', 'g.class2', 'g.class1', 'g.class0']
        """
        return list(self)

    def __iter__(self) -> Iterator[etree.Element]:
        """
        Make the selection as an iterator

        Returns
        -------
        Iterator[etree.Element]
            Iterator of non-None nodes
        """
        for group in self._groups:
            for node in group:
                if node is not None:
                    yield node

    def selection(self) -> Selection:
        """
        Returns the selection without any modification.

        Returns
        -------
        Selection
            Itself
        """
        return self

    def to_string(self, pretty_print: bool = True) -> str:
        """
        Convert selection to string

        Parameters
        ----------
        pretty_print : bool
            :code:`True` to prettify output

        Returns
        -------
        str
            String

        Examples
        --------

        >>> svg = d3.create("svg")
        >>> g = (
        ...     svg.select_all("g")
        ...     .data(list(reversed(range(10))))
        ...     .enter()
        ...     .append("g")
        ...     .attr("class", lambda d: f"class{d}")
        ... )
        >>> print(svg.to_string())
        <svg xmlns="http://www.w3.org/2000/svg">
          <g class="class9"/>
          <g class="class8"/>
          <g class="class7"/>
          <g class="class6"/>
          <g class="class5"/>
          <g class="class4"/>
          <g class="class3"/>
          <g class="class2"/>
          <g class="class1"/>
          <g class="class0"/>
        </svg>
        >>> print(svg.to_string(False))
        <svg xmlns="http://www.w3.org/2000/svg"><g class="class9"/><g class="class8"/><g class="class7"/><g class="class6"/><g class="class5"/><g class="class4"/><g class="class3"/><g class="class2"/><g class="class1"/><g class="class0"/></svg>
        >>> svg.to_string(False) == str(svg)
        True
        """
        if len(self._parents) == 0:
            return ""
        return (
            etree.tostring(self._parents[0], pretty_print=pretty_print)
            .decode("utf-8")
            .removesuffix("\n")
        )

    def to_repr(
        self, show_enter: bool = True, show_exit: bool = True, show_data: bool = True
    ) -> str:
        """
        Represents the selection with optional parameters.

        Parameters
        ----------
        show_enter : bool
            Show enter elements associated to this selection
        show_exit : bool
            Show exit elements associated to this selection
        show_data : bool
            Show data associated to this selection

        Returns
        -------
        str
            String

        Examples
        --------

        >>> svg = d3.create("svg")
        >>> g = (
        ...     svg.select_all("g")
        ...     .data(list(reversed(range(10))))
        ...     .enter()
        ...     .append("g")
        ...     .attr("class", lambda d: f"class{d}")
        ... )
        >>> print(g.to_repr())
        Selection(
            groups=[[g.class9, g.class8, g.class7, g.class6, g.class5, g.class4, g.class3, g.class2, g.class1, g.class0]],
            parents=[svg],
            enter=None,
            exit=None,
            data={<Element g at 0x7287cf850040>: 9, <Element g at 0x7287cfa50bc0>: 8, <Element g at 0x7287cf883ac0>: 7, <Element g at 0x7287cf8837c0>:6, <Element g at 0x7287cf883e00>: 5, <Element g at 0x7287cf883a40>: 4, <Element g at 0x7287cf882fc0>: 3, <Element g at 0x7287cf883a80>: 2, <Element g at 0x7287cf880d80>: 1, <Element g at 0x7287cf881000>: 0},
        )
        """

        def node_repr(node):
            if node is None:
                return str(node)
            if isinstance(node, EnterNode):
                return str(node)
            tag = node.tag
            class_name = node.attrib.get("class")
            if class_name:
                return f"{tag}.{class_name}"
            return tag

        groups = [
            f"[{', '.join(node_repr(node) for node in group)}]"
            for group in self._groups
        ]
        parents = f"[{', '.join(node_repr(parent) for parent in self._parents)}]"
        enter = f"    enter={self._enter},\n" if show_enter else ""
        exit = f"    exit={self._exit},\n" if show_exit else ""
        data = f"    data={self._data},\n" if show_data else ""
        return (
            f"Selection(\n    groups=[{', '.join(groups)}],\n    parents={parents},"
            f"\n{enter}{exit}{data})"
        )

    def __str__(self) -> str:
        """
        Returns the SVG content. Equivalent to
        :code:`Selection.to_string(False)`.

        Returns
        -------
        str
            String
        """
        return self.to_string(False)

    def __repr__(self) -> str:
        """
        Represents the selection

        Returns
        -------
        str
            String

        Examples
        --------

        >>> svg = d3.create("svg")
        >>> g = (
        ...     svg.select_all("g")
        ...     .data(list(reversed(range(10))))
        ...     .enter()
        ...     .append("g")
        ...     .attr("class", lambda d: f"class{d}")
        ... )
        >>> print(repr(g))
        Selection(
            groups=[[g.class9, g.class8, g.class7, g.class6, g.class5, g.class4, g.class3, g.class2, g.class1, g.class0]],
            parents=[svg],
        )
        """
        return self.to_repr(show_enter=False, show_exit=False, show_data=False)



def string_round(value, digit):
    value = str(round(float(value), digit)).removesuffix(".0")
    if "." in value:
        prefix, suffix = value.split(".")
        missing_zero = digit - len(suffix)
        if missing_zero > 0:
            suffix += "0" * missing_zero
        value = f"{prefix}.{suffix}"

    return "0" if value == "-0" else value



from __future__ import annotations

import math
from collections.abc import Callable
from typing import overload

from ..types import Number
from .continuous import Transformer, identity
from .init import init_range
from .linear import LinearBase
from .utils import as_float


def sign(x: float) -> float:
    return -1 if x < 0 else 1


def square(x: float) -> float:
    return sign(x) * x * x


def unsquare(x: float) -> float:
    return sign(x) * math.sqrt(abs(x))


class ScaleRadial(Transformer[float], LinearBase):
    """
    Radial scales are a variant of linear scales where the range
    is internally squared so that an input value corresponds linearly
    to the squared output value. These scales are useful when you want
    the input value to correspond to the area of a graphical mark and
    the mark is specified by radius, as in a radial bar chart.
    Radial scales do not support interpolate.

    Parameters
    ----------
    t : Callable[[float], float]
        Tranform function
    u : Callable[[float], float]
        Untranform function
    """

    def __init__(
        self,
        t: Callable[[float], float] = identity,
        u: Callable[[float], float] = identity,
    ):
        super().__init__(t, u)
        self._range_vals = [0, 1]
        self._round = False
        self._unknown = None

    def __call__(self, x: Number) -> Number:
        """
        Given a value from the domain, returns the corresponding value from the range.

        Parameters
        ----------
        x : Number
            Input value

        Returns
        -------
        Number
            Corresponding value from the range
        """
        y = unsquare(super().__call__(x))
        if isinstance(y, float) and math.isnan(y):
            return self._unknown
        elif self._round:
            return round(y)
        else:
            return y

    def invert(self, y: Number) -> Number:
        """
        Given a value from the range, returns the corresponding value
        from the domain. Inversion is useful for interaction, say to
        determine the data value corresponding to the position of the mouse.

        Parameters
        ----------
        y : Number
            Input value

        Returns
        -------
        Number
            Corresponding value from the domain
        """
        return super().invert(square(y))

    def set_range(self, range_vals: list[Number]) -> ScaleRadial:
        """
        Sets the scale's range to the specified array of values

        Parameters
        ----------
        range_vals : list[Number]
            Range values

        Returns
        -------
        ScaleRadial
            Itself
        """
        self._range_vals = [as_float(x) for x in range_vals]
        super().set_range([square(x) for x in self._range_vals])
        return self

    def get_range(self):
        return self._range_vals.copy()

    def set_range_round(self, range_vals: list[Number]) -> ScaleRadial:
        """
        Sets the scale's range to the specified array of values
        and sets scale's interpolator to :code:`interpolate_round`.

        Parameters
        ----------
        range_vals : list[Number]
            Range values

        Returns
        -------
        ScaleRadial
            Itself
        """
        return super().set_range(range_vals).set_round(True)

    def set_round(self, round_val: bool) -> ScaleRadial:
        """
        Enables or disables rounding accordingly

        Parameters
        ----------
        round_val : bool
            Round value

        Returns
        -------
        ScaleRadial
            Itself
        """
        self._round = bool(round_val)
        return self

    def get_round(self):
        return self._round

    def copy(self):
        return (
            ScaleRadial()
            .set_domain(self.get_domain())
            .set_range(self.get_range())
            .set_round(self.get_round())
            .set_clamp(self.get_clamp())
            .set_unknown(self.get_unknown())
        )


@overload
def scale_radial() -> ScaleRadial: ...


@overload
def scale_radial(range_vals: list[Number]) -> ScaleRadial: ...


@overload
def scale_radial(domain: list[Number], range_vals: list[Number]) -> ScaleRadial: ...


def scale_radial(*args):
    """
    Builds a new radial scale with the specified domain and range.

    Parameters
    ----------
    domain : list[Number]
        Array of numbers
    range_vals : list[Number]
        Array of values

    Returns
    -------
    ScaleRadial
        Scale object

    Examples
    --------

    >>> scale = d3.scale_radial([100, 200], [0, 480])
    >>> for x in range(11):
    ...     x = 100 + x * 10
    ...     print(x, scale(x))
    ...
    ...
    100 0.0
    110 151.7893276880822
    120 214.66252583997982
    130 262.9068276024797
    140 303.5786553761644
    150 339.4112549695428
    160 371.806401235912
    170 401.59681273635624
    180 429.32505167995964
    190 455.36798306424663
    200 480.0
    """
    scale = ScaleRadial()
    if len(args) == 1:
        return init_range(scale, range_vals=args[0])
    elif len(args) == 2:
        domain, range_vals = args
        return init_range(scale, domain=domain, range_vals=range_vals)
    return init_range(scale)



import math
from collections.abc import Callable

from ..array import argpass
from .exponent import exponent
from .format_group import format_group
from .format_numerals import format_numerals
from .format_prefix_auto import prefix_auto
from .format_specifier import FormatSpecifier, format_specifier
from .format_trim import format_trim
from .format_types import format_types
from .identity import identity

prefixes = [
    "y",
    "z",
    "a",
    "f",
    "p",
    "n",
    "µ",
    "m",
    "",
    "k",
    "M",
    "G",
    "T",
    "P",
    "E",
    "Z",
    "Y",
]


class Locale:
    def __init__(self, locale_def: dict):
        self.group = (
            format_group(locale_def["grouping"], locale_def["thousands"])
            if "grouping" in locale_def and "thousands" in locale_def
            else identity
        )
        self.currency_prefix = locale_def.get("currency", ["", ""])[0]
        self.currency_suffix = locale_def.get("currency", ["", ""])[1]
        self.decimal = locale_def.get("decimal", ".")
        self.numerals = (
            format_numerals(list(map(str, locale_def["numerals"])))
            if "numerals" in locale_def
            else identity
        )
        self.percent = locale_def.get("percent", "%")
        self.minus = locale_def.get("minus", "-")
        self.nan = locale_def.get("nan", "NaN")

    def format(self, specifier: str) -> Callable[[int | float], str]:
        """
        Returns a new format function for the given string specifier.
        The returned function takes a number as the only argument, and
        returns a string representing the formatted number.
        The general form of a specifier is:

        .. code::
            
            [[fill]align][sign][symbol][0][width][,][.precision][~][type]

        The :code:`fill` can be any character.
        The presence of a fill character is signaled by the
        align character following it, which must be one of the following:

        * :code:`>` - Forces the field to be right-aligned within the available space. (Default behavior).
        * :code:`<` - Forces the field to be left-aligned within the available space.
        * :code:`^` - Forces the field to be centered within the available space.
        * :code:`=` - like :code:`>`, but with any sign and symbol to the left of any padding.

        The :code:`sign` can be:

        * :code:`-` - nothing for zero or positive and a minus sign for negative. (Default behavior.)
        * :code:`+` - a plus sign for zero or positive and a minus sign for negative.
        * :code:`(` - nothing for zero or positive and parentheses for negative.
        * (space) - a space for zero or positive and a minus sign for negative.

        The :code:`symbol` can be:

        * :code:`$` - apply currency symbols per the locale definition.
        * :code:`#` - for binary, octal, or hexadecimal notation, \
        prefix by :code:`0b`, :code:`0o`, or :code:`0x`, respectively.

        The zero (:code:`0`) option enables zero-padding; this implicitly sets :code:`fill`
        to :code:`0` and align to :code:`=`. The width defines the minimum field width; if
        not specified, then the width will be determined by the content.
        The comma (:code:`,`) option enables the use of a group separator, such
        as a comma for thousands.


        The available :code:`type` values are:

        * :code:`e` - exponent notation.
        * :code:`f` - fixed point notation.
        * :code:`g` - either decimal or exponent notation, rounded to significant digits.
        * :code:`r` - decimal notation, rounded to significant digits.
        * :code:`s` - decimal notation with an SI prefix, rounded to significant digits.
        * :code:`%` - multiply by 100, and then decimal notation with a percent sign.
        * :code:`p` - multiply by 100, round to significant digits, and then decimal notation with a percent sign.
        * :code:`b` - binary notation, rounded to integer.
        * :code:`o` - octal notation, rounded to integer.
        * :code:`d` - decimal notation, rounded to integer.
        * :code:`x` - hexadecimal notation, using lower-case letters, rounded to integer.
        * :code:`X` - hexadecimal notation, using upper-case letters, rounded to integer.
        * :code:`c` - character data, for a string of text.

        Depending on the type, the precision either indicates the number of digits
        that follow the decimal point (types f and %), or the number of significant digits
        (types (space), :code:`e`, :code:`g`, :code:`r`, :code:`s` and :code:`p`). If the
        precision is not specified, it defaults to 6 for all types except (space) (none),
        which defaults to 12. Precision is ignored for integer formats (types :code:`b`,
        :code:`o`, :code:`d`, :code:`x`, and :code:`X`) and character data (type :code:`c`)

        The :code:`~` option trims insignificant trailing zeros across all format types.
        This is most commonly used in conjunction with types :code:`r`, :code:`e`, :code:`s` and :code:`%`.

        Parameters
        ----------
        specifier : str
            Specifier

        Returns
        -------
        Callable[[int | float], str]
            Format function

        Examples
        --------

        >>> d3.format(".0%")(0.123) # rounded percentage
        '12%'
        >>> d3.format("+20")(42) # space-filled and signed
        '                 +42'
        >>> d3.format(".^20")(42) # dot-filled and centered
        '.........42.........'
        >>> d3.format(".2s")(42e6) # SI-prefix with two significant digits
        '42M'
        >>> d3.format("#x")(48879) # prefixed lowercase hexadecimal
        '0xbeef'
        >>> d3.format(",.2r")(4223) # grouped thousands with two significant digits
        '4,200'
        >>> d3.format("s")(1500)
        '1.50000k'
        >>> d3.format("~s")(1500)
        '1.5k'
        >>> d3.format(".2")(42)
        '42'
        >>> d3.format(".2")(4.2)
        '4.2'
        >>> d3.format(".1")(42)
        '4e+01'
        >>> d3.format(".1")(4.2)
        '4'
        """
        if not isinstance(specifier, FormatSpecifier):
            specifier = format_specifier(specifier)
        fill = specifier.fill
        align = specifier.align
        sign = specifier.sign
        symbol = specifier.symbol
        zero = specifier.zero
        width = specifier.width
        comma = specifier.comma
        precision = specifier.precision
        trim = specifier.trim
        type_ = specifier.type

        if type_ == "n":
            comma = True
            type_ = "g"
        elif type_ not in format_types:
            if precision is None:
                precision = 12
            trim = True
            type_ = "g"

        if zero or (fill == "0" and align == "="):
            zero = True
            fill = "0"
            align = "="

        prefix = (
            self.currency_prefix
            if symbol == "$"
            else "0" + type_.lower()
            if symbol == "#" and type_.lower() in "box"
            else ""
        )
        suffix = (
            self.currency_suffix
            if symbol == "$"
            else self.percent
            if type_ in "%p"
            else ""
        )

        format_type = argpass(format_types[type_])
        maybe_suffix = type_ in "defgprs%"

        if precision is None:
            precision = 6
        elif type_ in "gprs":
            precision = max(1, min(21, precision))
        else:
            precision = max(0, min(20, precision))

        class Format:
            def __init__(self, prefix, suffix, group, numerals, minus, decimal):
                self.prefix = prefix
                self.suffix = suffix
                self.group = group
                self.numerals = numerals
                self.minus = minus
                self.decimal = decimal

            def __call__(self, value):
                value_prefix = self.prefix
                value_suffix = self.suffix

                if type_ == "c":
                    value_suffix = format_type(value, None) + self.suffix
                    value = ""
                else:
                    value = float(value)
                    value_negative = value < 0 or (value != 0 and 1 / value < 0)

                    value = (
                        self.nan
                        if math.isnan(value)
                        else format_type(abs(value), precision)
                    )

                    if trim:
                        value = format_trim(value)

                    if value_negative and sign != "+" and str(value) in ("0", "0."):
                        value_negative = False

                    if value_negative:
                        value_prefix = sign if sign == "(" else self.minus
                    elif sign == "-" or sign == "(":
                        value_prefix = ""
                    else:
                        value_prefix = sign

                    value_prefix += self.prefix

                    if type_ == "s":
                        value_suffix = prefixes[8 + prefix_auto.prefix_exponent // 3]
                    else:
                        value_suffix = ""

                    value_suffix += self.suffix
                    value_suffix += ")" if value_negative and sign == "(" else ""

                    if maybe_suffix:
                        for i, c in enumerate(value):
                            if not c.isdigit():
                                if c == ".":
                                    value_suffix = (
                                        self.decimal + value[i + 1 :] + value_suffix
                                    )
                                else:
                                    value_suffix = value[i:] + value_suffix
                                value = value[:i]
                                break

                if comma and not zero:
                    value = self.group(value, math.inf)

                length = len(value_prefix) + len(value) + len(value_suffix)
                padding = fill * (width - length) if width and length < width else ""

                if comma and zero:
                    value = self.group(
                        padding + value,
                        width - len(value_suffix) if len(padding) else math.inf,
                    )
                    padding = ""

                if align == "<":
                    value = value_prefix + value + value_suffix + padding
                elif align == "=":
                    value = value_prefix + padding + value + value_suffix
                elif align == "^":
                    length = len(padding) >> 1
                    value = (
                        padding[:length]
                        + value_prefix
                        + value
                        + value_suffix
                        + padding[length:]
                    )
                else:
                    value = padding + value_prefix + value + value_suffix

                return self.numerals(value)

            def __str__(self):
                return str(specifier)

        return Format(
            prefix, suffix, self.group, self.numerals, self.minus, self.decimal
        )

    def format_prefix(
        self, specifier: str, value: int | float
    ) -> Callable[[int | float], str]:
        """
        Equivalent to :code:`d3.format`, except the returned function will
        convert values to the units of the appropriate SI prefix for the
        specified numeric reference value before formatting in fixed point notation.
        The following prefixes are supported:

        * :code:`y` - yocto, :math:`10^{-24}`
        * :code:`z` - zepto, :math:`10^{-21}`
        * :code:`a` - atto, :math:`10^{-18}`
        * :code:`f` - femto, :math:`10^{-15}`
        * :code:`p` - pico, :math:`10^{-12}`
        * :code:`n` - nano, :math:`10^{-9}`
        * :code:`µ` - micro, :math:`10^{-6}`
        * :code:`m` - milli, :math:`10^{-3}`
        * (none) - :math:`1`
        * :code:`k` - kilo, :math:`10^{3}`
        * :code:`M` - mega, :math:`10^{6}`
        * :code:`G` - giga, :math:`10^{9}`
        * :code:`T` - tera, :math:`10^{12}`
        * :code:`P` - peta, :math:`10^{15}`
        * :code:`E` - exa, :math:`10^{18}`
        * :code:`Z` - zetta, :math:`10^{21}`
        * :code:`Y` - yotta, :math:`10^{24}`

        Parameters
        ----------
        specifier : str
            Specifier
        value : int | float
            Numeric reference

        Returns
        -------
        Callable[[int | float], str]
            Format function

        Examples
        --------
        >>> f = d3.format_prefix(",.0", 1e-6)
        >>> f(0.00042)
        '420µ'
        >>> f(0.0042)
        '4,200µ'
        """
        if not isinstance(specifier, FormatSpecifier):
            specifier = format_specifier(specifier)
        specifier.type = "f"
        f = self.format(specifier)
        e = max(-8, min(8, exponent(value) // 3)) * 3
        k = 10**-e
        prefix = prefixes[8 + e // 3]

        def format_with_prefix(value):
            return f(k * value) + prefix

        return format_with_prefix



from __future__ import annotations

from collections.abc import Callable
from math import isnan

from ..array import argpass
from ..types import SimulationNode, SimulationNodeFunction
from .constant import constant


class ForceX:
    def __init__(self, x: SimulationNodeFunction[float]):
        self._x = x
        self._strength = constant(0.1)
        self._nodes = None
        self._strengths = None
        self._xz = None

    def __call__(self, alpha: float):
        for i, node in enumerate(self._nodes):
            node["vx"] += (self._xz[i] - node["x"]) * self._strengths[i] * alpha

    def _initialize(self):
        if self._nodes is None:
            return

        self._strengths = [None] * len(self._nodes)
        self._xz = [None] * len(self._nodes)
        for i, node in enumerate(self._nodes):
            self._xz[i] = x = self._x(node, i, self._nodes)
            self._strengths[i] = 0 if isnan(x) else self._strength(node, i, self._nodes)

    def initialize(self, nodes: list[SimulationNode], random: Callable[[None], float]):
        self._nodes = nodes
        self._initialize()

    def set_strength(self, strength: SimulationNodeFunction[float] | float) -> ForceX:
        """
        Sets the strength accessor to the specified number or function,
        re-evaluates the strength accessor for each node, and returns this
        force.

        The strength determines how much to increment the node's x-velocity:
        :code:`(x - node["x"]) * strength`. For example, a value of :math:`0.1`
        indicates that the node should move a tenth of the way from its current
        x-position to the target x-position with each application. Higher
        values moves nodes more quickly to the target position, often at the
        expense of other forces or constraints. A value outside the range
        :math:`[0,1]` is not recommended.

        Parameters
        ----------
        strength : SimulationNodeFunction[float] | float
            Strength function or constant value. If it is a function, it takes
            the following arguments:

            * **node** (:code:`SimulationNode`) - the node element
            * **i** (:code:`int`) - the index of the node
            * **nodes** (:code:`list[SimulationNode]`) - the list of nodes

            It returns the strength value (:code:`float`)

        Returns
        -------
        ForceX
            Itself
        """
        if callable(strength):
            self._strength = argpass(strength)
        else:
            self._strength = constant(strength)
        self._initialize()
        return self

    def x(self, x: SimulationNodeFunction[float] | float) -> ForceX:
        """
        Sets the x-coordinate accessor to the specified number or function,
        re-evaluates the x-accessor for each node, and returns this force.

        Parameters
        ----------
        x : SimulationNodeFunction[float] | float
            x-coordinate function or constant value. If it is a function, it takes
            the following arguments:

            * **node** (:code:`SimulationNode`) - the node element
            * **i** (:code:`int`) - the index of the node
            * **nodes** (:code:`list[SimulationNode]`) - the list of nodes

            It returns the x-coordinate value (:code:`float`)

        Returns
        -------
        ForceX
            Itself
        """
        if callable(x):
            self._x = argpass(x)
        else:
            self._x = constant(x)
        self._initialize()
        return self

    def get_strength(self) -> SimulationNodeFunction[float]:
        return self._strength

    def get_x(self) -> SimulationNodeFunction[float]:
        return self._x


def force_x(x: SimulationNodeFunction[float] | float | None = None) -> ForceX:
    """
    The x-position force pushes nodes towards a desired position along the
    given dimension with a configurable strength.

    The strength of the force is proportional to the one-dimensional distance
    between the node's position and the target position. While these forces can
    be used to position individual nodes, they are intended primarily for
    global forces that apply to all (or most) nodes.

    Parameters
    ----------
    x : SimulationNodeFunction[float] | float | None
        x-coordinate function or constant value. If it is a function, it takes
        the following arguments:

        * **node** (:code:`SimulationNode`) - the node element
        * **i** (:code:`int`) - the index of the node
        * **nodes** (:code:`list[SimulationNode]`) - the list of nodes

        It returns the x-coordinate value (:code:`float`)

    Returns
    -------
    ForceX
        Force object
    """
    if not callable(x):
        x = constant(0 if x is None else x)
    else:
        x = argpass(x)

    return ForceX(argpass(x))



from math import sqrt

TAN30 = sqrt(1 / 3)
TAN30_2 = TAN30 * 2


def symbol_diamond(context, size):
    y = sqrt(size / TAN30_2)
    x = y * TAN30
    context.move_to(0, -y)
    context.line_to(x, 0)
    context.line_to(0, y)
    context.line_to(-x, 0)
    context.close_path()



from math import floor

from ...color import cubehelix
from ...interpolate import interpolate_cubehelix_long

interpolate_warm = interpolate_cubehelix_long(
    cubehelix(-100, 0.75, 0.35), cubehelix(80, 1.50, 0.8)
)
interpolate_cool = interpolate_cubehelix_long(
    cubehelix(260, 0.75, 0.35), cubehelix(80, 1.50, 0.8)
)


def interpolate_rainbow(t):
    if t < 0 or t > 1:
        t -= floor(t)
    ts = abs(t - 0.5)
    c = cubehelix(360 * t - 100, 1.5 - 1.5 * ts, 0.8 - 0.9 * ts)
    return str(c)



from ..colors import colors
from ..ramp import ramp

scheme = [
    "e0ecf49ebcda8856a7",
    "edf8fbb3cde38c96c688419d",
    "edf8fbb3cde38c96c68856a7810f7c",
    "edf8fbbfd3e69ebcda8c96c68856a7810f7c",
    "edf8fbbfd3e69ebcda8c96c68c6bb188419d6e016b",
    "f7fcfde0ecf4bfd3e69ebcda8c96c68c6bb188419d6e016b",
    "f7fcfde0ecf4bfd3e69ebcda8c96c68c6bb188419d810f7c4d004b",
]
SCHEME_BUPU = list(map(colors, scheme))
interpolate_bupu = ramp(SCHEME_BUPU)



from __future__ import annotations

from collections.abc import Callable
from typing import Literal

from ..array import argpass
from ..selection.selection import Selection
from ..types import ContinuousScaler, SequentialScaler

TOP = 1
RIGHT = 2
BOTTOM = 3
LEFT = 4
EPSILON = 1e-6


def translate_x(x):
    return f"translate({x}, 0)"


def translate_y(y):
    return f"translate(0, {y})"


def number(scale):
    def f(d):
        return float(scale(d))

    return f


def center(scale, offset):
    offset = max(0, scale.get_bandwidth() - offset * 2) / 2
    if scale.get_round():
        offset = round(offset)
    return lambda d: float(scale(d)) + offset


# def entering(context):
#     return not hasattr(context, "__axis")


class Axis:
    """
    Builds a new oriented axis generator for the given scale,
    with empty tick arguments, a tick size of 6 and padding of 3.

    Parameters
    ----------
    orient : Literal[1, 2, 3, 4]
        Orientation
    scale : ContinuousScaler | SequentialScaler
        Scaler
    """

    def __init__(
        self,
        orient: Literal[1, 2, 3, 4],
        scale: ContinuousScaler | SequentialScaler,
    ):
        self._scale = scale
        self._orient = orient
        self._tick_arguments = []
        self._tick_values = None
        self._tick_format = None
        self._tick_size_inner = 6
        self._tick_size_outer = 6
        self._tick_padding = 3
        self._offset = 0.5
        self._k = -1 if orient in [TOP, LEFT] else 1
        self._x = "x" if orient in [LEFT, RIGHT] else "y"
        self._transform = translate_x if orient in [TOP, BOTTOM] else translate_y

    def __call__(self, context: Selection):
        """
        Render the axis to the given context, which may be either a selection
        of SVG containers (either SVG or G elements) or a corresponding transition.

        Parameters
        ----------
        context : Selection
            Context

        Examples
        --------

        >>> import detroit as d3
        >>> svg = d3.create("svg")
        >>> svg.call(d3.axis_bottom(d3.scale_linear()))
        Selection(
            groups=[[svg]],
            parents=[svg],
        )
        >>> print(svg.to_string())
        <svg xmlns="http://www.w3.org/2000/svg" fill="none" font-size="10" font-family="sans-serif" text-anchor="middle">
          <path class="domain" stroke="currentColor" d="M0.5,6V0.5H1.5V6"/>
          <g class="tick" opacity="1" transform="translate(0.5, 0)">
            <line stroke="currentColor" y2="6"/>
            <text fill="currentColor" y="9" dy="0.71em">0.0</text>
          </g>
          <g class="tick" opacity="1" transform="translate(0.6, 0)">
            <line stroke="currentColor" y2="6"/>
            <text fill="currentColor" y="9" dy="0.71em">0.1</text>
          </g>
          <g class="tick" opacity="1" transform="translate(0.7, 0)">
            <line stroke="currentColor" y2="6"/>
            <text fill="currentColor" y="9" dy="0.71em">0.2</text>
          </g>
          <g class="tick" opacity="1" transform="translate(0.8, 0)">
            <line stroke="currentColor" y2="6"/>
            <text fill="currentColor" y="9" dy="0.71em">0.3</text>
          </g>
          <g class="tick" opacity="1" transform="translate(0.9, 0)">
            <line stroke="currentColor" y2="6"/>
            <text fill="currentColor" y="9" dy="0.71em">0.4</text>
          </g>
          <g class="tick" opacity="1" transform="translate(1.0, 0)">
            <line stroke="currentColor" y2="6"/>
            <text fill="currentColor" y="9" dy="0.71em">0.5</text>
          </g>
          <g class="tick" opacity="1" transform="translate(1.1, 0)">
            <line stroke="currentColor" y2="6"/>
            <text fill="currentColor" y="9" dy="0.71em">0.6</text>
          </g>
          <g class="tick" opacity="1" transform="translate(1.2, 0)">
            <line stroke="currentColor" y2="6"/>
            <text fill="currentColor" y="9" dy="0.71em">0.7</text>
          </g>
          <g class="tick" opacity="1" transform="translate(1.3, 0)">
            <line stroke="currentColor" y2="6"/>
            <text fill="currentColor" y="9" dy="0.71em">0.8</text>
          </g>
          <g class="tick" opacity="1" transform="translate(1.4, 0)">
            <line stroke="currentColor" y2="6"/>
            <text fill="currentColor" y="9" dy="0.71em">0.9</text>
          </g>
          <g class="tick" opacity="1" transform="translate(1.5, 0)">
            <line stroke="currentColor" y2="6"/>
            <text fill="currentColor" y="9" dy="0.71em">1.0</text>
          </g>
        </svg>

        Notes
        -----

        Transitions are not yet implemented
        """
        if self._tick_values is not None:
            values = self._tick_values
        elif hasattr(self._scale, "ticks"):
            values = argpass(self._scale.ticks)(*self._tick_arguments)
        else:
            values = self._scale.get_domain()

        if self._tick_format is not None:
            format_func = self._tick_format
        elif hasattr(self._scale, "tick_format"):
            format_func = argpass(self._scale.tick_format)(*self._tick_arguments)
        else:

            def format_func(d):
                return d

        spacing = max(self._tick_size_inner, 0) + self._tick_padding
        range_values = self._scale.get_range()
        range0 = float(range_values[0]) + self._offset
        range1 = float(range_values[-1]) + self._offset

        if hasattr(self._scale, "get_bandwidth"):
            position = center(self._scale.copy(), self._offset)
        else:
            position = number(self._scale.copy())

        selection = context.selection() if hasattr(context, "selection") else context
        path = selection.select_all(".domain").data([None])
        tick = selection.select_all(".tick").data(values, self._scale).order()
        tick_exit = tick.exit()
        tick_enter = tick.enter().append("g").attr("class", "tick")
        line = tick.select("line")
        text = tick.select("text")

        path = path.merge(
            path.enter()
            .insert("path", ".tick")
            .attr("class", "domain")
            .attr("stroke", "currentColor")
        )

        tick = tick.merge(tick_enter)

        line = line.merge(
            tick_enter.append("line")
            .attr("stroke", "currentColor")
            .attr(f"{self._x}2", self._k * self._tick_size_inner)
        )

        text = text.merge(
            tick_enter.append("text")
            .attr("fill", "currentColor")
            .attr(self._x, self._k * spacing)
            .attr(
                "dy",
                "0em"
                if self._orient == TOP
                else "0.71em"
                if self._orient == BOTTOM
                else "0.32em",
            )
        )

        # TODO : transition implementation
        # if context != selection:
        #     path = path.transition(context)
        #     tick = tick.transition(context)
        #     line = line.transition(context)
        #     text = text.transition(context)
        #
        #     def transform(d):
        #         d = position(d)
        #         if d is not None and math.isfinite(d):
        #             return self._transform(position(d) + self._offset)
        #
        #     tick_exit = (
        #         tick_exit.transition(context)
        #                  .attr("opacity", EPSILON)
        #                  .attr("transform", transform)
        #     )
        #
        #     def transform(d): # TODO
        #         return None
        #
        #     tick_enter.attr("opacity", EPSILON).attr("transform", transform)

        tick_exit.remove()

        if self._orient == LEFT or self._orient == RIGHT:
            if self._tick_size_outer:
                path.attr(
                    "d",
                    f"M{self._k * self._tick_size_outer},{range0}H{self._offset}V{range1}H{self._k * self._tick_size_outer}",
                )
            else:
                path.attr("d", f"M{self._offset},{range0}V{range1}")
        else:
            if self._tick_size_outer:
                path.attr(
                    "d",
                    f"M{range0},{self._k * self._tick_size_outer}V{self._offset}H{range1}V{self._k * self._tick_size_outer}",
                )
            else:
                path.attr("d", f"M{range0},{self._offset}H{range1}")

        def transform(d, *args):
            return self._transform(position(d) + self._offset)

        tick.attr("opacity", 1).attr("transform", transform)

        line.attr(f"{self._x}2", self._k * self._tick_size_inner)

        text.attr(self._x, self._k * spacing).text(format_func)

        (
            selection.attr("fill", "none")  # .filter(entering(context))
            .attr("font-size", 10)
            .attr("font-family", "sans-serif")
            .attr(
                "text-anchor",
                "start"
                if self._orient == RIGHT
                else "end"
                if self._orient == LEFT
                else "middle",
            )
        )

    def set_scale(self, scale: ContinuousScaler | SequentialScaler) -> Axis:
        """
        Sets scale value

        Parameters
        ----------
        scale : ContinuousScaler | SequentialScaler
            New scale

        Returns
        -------
        Axis
            Itself

        Examples
        --------

        >>> scale = d3.scale_band()
        >>> x = d3.axis_bottom(d3.scale_linear())
        >>> print(x)
        Axis[BOTTOM](
            scale=ScaleLinear(domain=[0, 1], range=[0, 1]),
            tick_arguments=[],
            tick_values=None,
            tick_size_inner=6,
            tick_size_outer=6,
            tick_padding=3,
            offset=0.5,
        )
        >>> x.set_scale(scale)
        <Axis[BOTTOM] at 0x7fb4e098e9b0>
        >>> print(x)
        Axis[BOTTOM](
            scale=ScaleBand(domain=[], range=[0, 1], padding_inner=0, padding_outer=0),
            tick_arguments=[],
            tick_values=None,
            tick_size_inner=6,
            tick_size_outer=6,
            tick_padding=3,
            offset=0.5,
        )
        """
        self._scale = scale
        return self

    def set_ticks(self, *ticks: int | str) -> Axis:
        """
        Tick values will be passed to :code:`scale.ticks` and
        :code:`scale.tick_format` when :code:`Axis.__call__` is called

        Parameters
        ----------
        *ticks : int | str
            It depends on :code:`scale` type. Most of the time, it is the
            *count* for the number of ticks and optional format specifier

        Returns
        -------
        Axis
            Itself

        Examples
        --------

        To generate three ticks with percentage formatting on a linear scale :

        >>> width = 810
        >>> height = 30
        >>> margin = 15
        >>> svg = (
        ...     d3.create("svg")
        ...     .attr("width", width)
        ...     .attr("height", height)
        ...     .attr("viewBox", f"0 0 {width} {height}")
        ... )
        >>> axis = d3.axis_bottom(d3.scale_linear().set_range([0, width - 2 * margin]))
        >>> svg.append("g").attr("transform", f"translate({margin}, {margin * 0.5})").call(
        ...     axis.set_ticks(3, "%")
        ... )
        Selection(
            groups=[[g]],
            parents=[svg],
        )

        **Result**

        .. image:: ../figures/light_axis_ticks_1.svg
           :align: center
           :class: only-light

        .. image:: ../figures/dark_axis_ticks_1.svg
           :align: center
           :class: only-dark

        ------------

        To generate ticks every fifteen minutes with a time scale :

        >>> from datetime import datetime
        >>> width = 810
        >>> height = 30
        >>> margin = 15
        >>> svg = (
        ...     d3.create("svg")
        ...     .attr("width", width)
        ...     .attr("height", height)
        ...     .attr("viewBox", f"0 0 {width} {height}")
        ... )
        >>> scale = d3.scale_time(
        ...     [datetime(2000, 1, 1, 12, 0), datetime(2000, 1, 1, 13, 0)], [0, 1]
        ... )
        >>> axis = d3.axis_bottom(scale.set_range([0, width - 2 * margin]))
        >>> svg.append("g").attr("transform", f"translate({margin}, {margin * 0.5})").call(
        ...     axis.set_ticks(d3.time_minute.every(15))
        ... )
        Selection(
            groups=[[g]],
            parents=[svg],
        )

        **Result**

        .. image:: ../figures/light_axis_ticks_2.svg
           :align: center
           :class: only-light

        .. image:: ../figures/dark_axis_ticks_2.svg
           :align: center
           :class: only-dark

        ------------

        Note the similarities with :code:`set_tick_arguments`:

        >>> axis.set_ticks(10) # same as axis.set_tick_arguments([10])
        """
        self._tick_arguments = list(ticks)
        return self

    def set_tick_arguments(self, tick_arguments: list[int | str]) -> Axis:
        """
        Tick arguments will be passed to :code:`scale.ticks` and
        :code:`scale.tick_format` when :code:`Axis.__call__` is called

        Parameters
        ----------
        tick_arguments : list[int | str]
            It depends on :code:`scale` type. Most of the time, it is the
            *count* for the number of ticks and optional format specifier

        Returns
        -------
            Itself

        Examples
        --------

        To generate three ticks with percentage formatting on a linear scale :

        >>> # same as axis.set_ticks(3, "%")
        >>> axis.set_tick_arguments([3, "%"])

        To generate ticks every fifteen minutes with a time scale :

        >>> # same as axis.set_ticks(d3.time_minute.every(15))
        >>> axis.set_tick_arguments([d3.time_minute.every(15)])

        Notes
        -----

        Check :code:`Axis.set_ticks` for more details in examples
        """
        self._tick_arguments = list(tick_arguments)
        return self

    def set_tick_values(self, tick_values: list[int | float]) -> Axis:
        """
        Ticks values are used for ticks rather than the scale’s
        automatic tick generator.

        Parameters
        ----------
        tick_values : list[int | float]
            Tick values

        Returns
        -------
        Axis
            Itself

        Examples
        --------

        >>> width = 810
        >>> height = 30
        >>> margin = 15
        >>> svg = (
        ...     d3.create("svg")
        ...     .attr("width", width)
        ...     .attr("height", height)
        ...     .attr("viewBox", f"0 0 {width} {height}")
        ... )
        >>> axis = d3.axis_bottom(
        ...     d3.scale_linear().set_domain([0, 360]).set_range([0, width - 2 * margin])
        ... )
        >>> (
        ...     svg.append("g")
        ...     .attr("transform", f"translate({margin}, {margin * 0.5})")
        ...     .call(axis.set_tick_values([0, 120, 240, 360]))
        ... )
        Selection(
            groups=[[g]],
            parents=[svg],
        )

        **Result**

        .. image:: ../figures/light_axis_angle.svg
           :align: center
           :class: only-light

        .. image:: ../figures/dark_axis_angle.svg
           :align: center
           :class: only-dark
        """
        self._tick_values = tick_values
        return self

    def set_tick_format(self, tick_format: Callable[[int | float], str]) -> Axis:
        """
        Sets the tick format function and returns the axis.

        Parameters
        ----------
        tick_format : Callable[[int | float], str]
            Tick formatter

        Returns
        -------
        Axis
            Itself

        Examples
        --------

        For example, to display integers as decimal notation rounded to
        significant digits:

        >>> width = 810
        >>> height = 30
        >>> margin = 15
        >>> svg = (
        ...     d3.create("svg")
        ...     .attr("width", width)
        ...     .attr("height", height)
        ...     .attr("viewBox", f"0 0 {width} {height}")
        ... )
        >>> axis = d3.axis_bottom(
        ...     d3.scale_linear().set_domain([0, 10_000]).set_range([0, width - 2 * margin])
        ... )
        >>> (
        ...     svg.append("g")
        ...     .attr("transform", f"translate({margin}, {margin * 0.5})")
        ...     .call(axis.set_tick_format(d3.format("~s")))
        ... )
        Selection(
            groups=[[g]],
            parents=[svg],
        )

        **Result**

        .. image:: ../figures/light_axis_format.svg
           :align: center
           :class: only-light

        .. image:: ../figures/dark_axis_format.svg
           :align: center
           :class: only-dark
        """
        self._tick_format = tick_format
        return self

    def set_tick_size(self, tick_size: int) -> Axis:
        """
        Sets the inner and outer tick size to the specified
        value and returns the axis.

        Parameters
        ----------
        tick_size : int
            Tick size

        Returns
        -------
        Axis
            Itself

        Examples
        --------

        >>> width = 810
        >>> height = 30
        >>> margin = 15
        >>> svg = (
        ...     d3.create("svg")
        ...     .attr("width", width)
        ...     .attr("height", height)
        ...     .attr("viewBox", f"0 0 {width} {height}")
        ... )
        >>> axis = d3.axis_bottom(d3.scale_linear().set_range([0, width - 2 * margin]))
        >>> (
        ...     svg.append("g")
        ...     .attr("transform", f"translate({margin}, {margin * 0.5})")
        ...     .call(axis.set_tick_size(12))
        ... )
        Selection(
            groups=[[g]],
            parents=[svg],
        )

        **Result**

        .. image:: ../figures/light_axis_tick_size.svg
           :align: center
           :class: only-light

        .. image:: ../figures/dark_axis_tick_size.svg
           :align: center
           :class: only-dark
        """
        self._tick_size_inner = self._tick_size_outer = tick_size
        return self

    def set_tick_size_inner(self, tick_size_inner: int) -> Axis:
        """
        Sets the inner tick size to the specified value and
        returns the axis.

        Parameters
        ----------
        tick_size_inner : int
            Inner tick size

        Returns
        -------
        Axis
            Itself

        Examples
        --------

        >>> width = 810
        >>> height = 30
        >>> margin = 15
        >>> svg = (
        ...     d3.create("svg")
        ...     .attr("width", width)
        ...     .attr("height", height)
        ...     .attr("viewBox", f"0 0 {width} {height}")
        ... )
        >>> axis = d3.axis_bottom(d3.scale_linear().set_range([0, width - 2 * margin]))
        >>> (
        ...     svg.append("g")
        ...     .attr("transform", f"translate({margin}, {margin * 0.5})")
        ...     .call(axis.set_tick_size_inner(0))
        ... )
        Selection(
            groups=[[g]],
            parents=[svg],
        )

        **Result**

        .. image:: ../figures/light_axis_tick_size_inner.svg
           :align: center
           :class: only-light

        .. image:: ../figures/dark_axis_tick_size_inner.svg
           :align: center
           :class: only-dark

        Notes
        -----

        Inner ticks are all ticks with an associated value.
        """
        self._tick_size_inner = tick_size_inner
        return self

    def set_tick_size_outer(self, tick_size_outer: int) -> Axis:
        """
        Sets the outer tick size to the specified value and
        returns the axis.

        Parameters
        ----------
        tick_size_outer : int
            Outer tick size

        Returns
        -------
        Axis
            Itself

        Examples
        --------

        >>> width = 810
        >>> height = 30
        >>> margin = 15
        >>> svg = (
        ...     d3.create("svg")
        ...     .attr("width", width)
        ...     .attr("height", height)
        ...     .attr("viewBox", f"0 0 {width} {height}")
        ... )
        >>> axis = d3.axis_bottom(d3.scale_linear().set_range([0, width - 2 * margin]))
        >>> (
        ...     svg.append("g")
        ...     .attr("transform", f"translate({margin}, {margin * 0.5})")
        ...     .call(axis.set_tick_size_outer(12))
        ... )
        Selection(
            groups=[[g]],
            parents=[svg],
        )

        **Result**

        .. image:: ../figures/light_axis_tick_size_outer.svg
           :align: center
           :class: only-light

        .. image:: ../figures/dark_axis_tick_size_outer.svg
           :align: center
           :class: only-dark

        Notes
        -----

        Outer ticks are only ticks on extremities of the axis.
        """
        self._tick_size_outer = tick_size_outer
        return self

    def set_tick_padding(self, tick_padding: int) -> Axis:
        """
        Sets the padding to the specified value in pixels.

        Parameters
        ----------
        tick_padding : int
            Tick padding in pixels

        Returns
        -------
        Axis
            Itself

        Examples
        --------

        >>> width = 810
        >>> height = 30
        >>> margin = 15
        >>> svg = (
        ...     d3.create("svg")
        ...     .attr("width", width)
        ...     .attr("height", height)
        ...     .attr("viewBox", f"0 0 {width} {height}")
        ... )
        >>> axis = d3.axis_bottom(d3.scale_linear().set_range([0, width - 2 * margin]))
        >>> (
        ...     svg.append("g")
        ...     .attr("transform", f"translate({margin}, {margin * 0.5})")
        ...     .call(axis.set_tick_padding(8))
        ... )
        Selection(
            groups=[[g]],
            parents=[svg],
        )

        **Result**

        .. image:: ../figures/light_axis_tick_padding.svg
           :align: center
           :class: only-light

        .. image:: ../figures/dark_axis_tick_padding.svg
           :align: center
           :class: only-dark

        Notes
        -----

        The value :code:`y` is computing as :code:`max(self._tick_size_inner,
        0) + self._tick_padding`.
        """
        self._tick_padding = tick_padding
        return self

    def set_offset(self, offset: int | float) -> Axis:
        """
        Sets the pixel offset to the specified value in pixels.

        Parameters
        ----------
        offset : int | float
            Pixel offset

        Returns
        -------
        Axis
            Itself

        Examples
        --------

        >>> width = 810
        >>> height = 30
        >>> margin = 15
        >>> svg = (
        ...     d3.create("svg")
        ...     .attr("width", width)
        ...     .attr("height", height)
        ...     .attr("viewBox", f"0 0 {width} {height}")
        ... )
        >>> axis = d3.axis_bottom(d3.scale_linear().set_range([0, width - 2 * margin]))
        >>> (
        ...     svg.append("g")
        ...     .attr("transform", f"translate({margin}, {margin * 0.5})")
        ...     .call(axis.set_offset(8))
        ... )
        Selection(
            groups=[[g]],
            parents=[svg],
        )

        **Result**

        .. image:: ../figures/light_axis_offset.svg
           :align: center
           :class: only-light

        .. image:: ../figures/dark_axis_offset.svg
           :align: center
           :class: only-dark
        """
        self._offset = offset
        return self

    def get_scale(self) -> ContinuousScaler | SequentialScaler:
        return self._scale

    def get_tick_arguments(self) -> list[int | str]:
        return self._tick_arguments.copy()

    def get_tick_values(self) -> list[int | float]:
        return self._tick_values

    def get_tick_format(self) -> Callable[[int | float], str]:
        return self._tick_format

    def get_tick_size_inner(self) -> float:
        return self._tick_size_inner

    def get_tick_size_outer(self) -> float:
        return self._tick_size_outer

    def get_tick_padding(self) -> float:
        return self._tick_padding

    def get_offset(self) -> float:
        return self._offset

    def __str__(self) -> str:
        orients = ["TOP", "RIGHT", "BOTTOM", "LEFT"]
        name = f"Axis[{orients[self._orient - 1]}]"
        scale = str(self.get_scale()).split("\n")
        scale = "\n    ".join(scale)
        attrbs = [
            "tick_arguments",
            "tick_values",
            "tick_size_inner",
            "tick_size_outer",
            "tick_padding",
            "offset",
        ]
        attrbs = (f"{a}={getattr(self, f'get_{a}')()}," for a in attrbs)
        attrbs = "\n    ".join(attrbs)
        return f"{name}(\n    scale={scale},\n    {attrbs}\n)"

    def __repr__(self) -> str:
        orients = ["TOP", "RIGHT", "BOTTOM", "LEFT"]
        name = f"Axis[{orients[self._orient - 1]}]"
        addr = id(self)
        return f"<{name} at {hex(addr)}>"


def axis_top(scale: ContinuousScaler | SequentialScaler) -> Axis:
    """
    Builds a new top-oriented axis generator for the given scale,
    with empty tick arguments, a tick size of 6 and padding of 3.
    In this orientation, ticks are drawn above the horizontal domain path.

    Parameters
    ----------
    scale : ContinuousScaler | SequentialScaler
        Scaler

    Returns
    -------
    Axis
        Axis
    """
    return Axis(TOP, scale)


def axis_right(scale: ContinuousScaler | SequentialScaler) -> Axis:
    """
    Builds a new right-oriented axis generator for the given scale,
    with empty tick arguments, a tick size of 6 and padding of 3.
    In this orientation, ticks are drawn above the vertical domain path.

    Parameters
    ----------
    scale : ContinuousScaler | SequentialScaler
        Scaler

    Returns
    -------
    Axis
        Axis
    """
    return Axis(RIGHT, scale)


def axis_bottom(scale: ContinuousScaler | SequentialScaler) -> Axis:
    """
    Builds a new bottom-oriented axis generator for the given scale,
    with empty tick arguments, a tick size of 6 and padding of 3.
    In this orientation, ticks are drawn above the horizontal domain path.

    Parameters
    ----------
    scale : ContinuousScaler | SequentialScaler
        Scaler

    Returns
    -------
    Axis
        Axis
    """
    return Axis(BOTTOM, scale)


def axis_left(scale: ContinuousScaler | SequentialScaler) -> Axis:
    """
    Builds a new left-oriented axis generator for the given scale,
    with empty tick arguments, a tick size of 6 and padding of 3.
    In this orientation, ticks are drawn above the vertical domain path.

    Parameters
    ----------
    scale : ContinuousScaler | SequentialScaler
        Scaler

    Returns
    -------
    Axis
        Axis
    """
    return Axis(LEFT, scale)



from math import nan, pi

from ..common import Context, PolygonStream

TAU = 2 * pi


class PathContext(PolygonStream):
    def __init__(self, context: Context):
        self._context = context
        self._radius = 4.5
        self._line = nan
        self._point = nan

    def point_radius(self, radius: float):
        self._radius = radius
        return self

    def polygon_start(self):
        self._line = 0

    def polygon_end(self):
        self._line = nan

    def line_start(self):
        self._point = 0

    def line_end(self):
        if self._line == 0:
            self._context.close_path()
        self._point = nan

    def point(self, x: float, y: float):
        if self._point == 0:
            self._context.move_to(x, y)
            self._point = 1
        elif self._point == 1:
            self._context.line_to(x, y)
        else:
            self._context.move_to(x + self._radius, y)
            self._context.arc(x, y, self._radius, 0, TAU)

    def result(self):
        return



from collections.abc import Callable
from datetime import datetime

from ..color.color import Color, color
from ..types import T, U, V
from .constant import constant
from .date import interpolate_date
from .number import interpolate_number
from .number_array import interpolate_number_array, is_number_array
from .rgb import interpolate_rgb
from .string import interpolate_string


def interpolate_object(a: dict[U, V], b: dict[U, V]) -> Callable[[float], dict[U, V]]:
    """
    Returns an interpolator between the two objects a and b.

    Parameters
    ----------
    a : dict[U, V]
        Object a
    b : dict[U, V]
        Object b

    Returns
    -------
    Callable[[float], dict[U, V]]
        Interpolator function

    Examples
    --------

    >>> interpolator = d3.interpolate_object({"a": [0, 10, 20]}, {"a": [30, 40, 50]})
    >>> interpolator(0)
    {'a': [0, 10, 20]}
    >>> interpolator(1)
    {'a': [30, 40, 50]}
    >>> interpolator(0.5)
    {'a': [15.0, 25.0, 35.0]}
    """
    i = {}
    c = {}

    if a is None or not isinstance(a, dict):
        a = {}
    if b is None or not isinstance(b, dict):
        b = {}

    for k in b:
        if k in a:
            i[k] = interpolate(a.get(k), b.get(k))
        else:
            c[k] = b.get(k)

    def local_interpolate(t):
        for k in i:
            c[k] = i[k](t)
        return c

    return local_interpolate


def interpolate_array(a: list[T], b: list[T]) -> Callable[[float], list[T]]:
    """
    Returns an interpolator between the two arrays a and b.

    Parameters
    ----------
    a : list[T]
        Array a
    b : list[T]
        Array b

    Returns
    -------
    Callable[[float], list[T]]
        Interpolator function

    Examples
    --------

    >>> interpolator = d3.interpolate_array([0, 10, 20], [30, 40, 50])
    >>> interpolator(0)
    [0, 10, 20]
    >>> interpolator(1)
    [30, 40, 50]
    >>> interpolator(0.5)
    [15.0, 25.0, 35.0]
    """
    return interpolate_number_array(a, b) if is_number_array(b) else generic_array(a, b)


def generic_array(a: list[T], b: list[T]) -> Callable[[float], T]:
    """
    Returns an interpolator between two sequences a and b where stored values
    are interpolated recursively.

    Parameters
    ----------
    a : list[T]
        a sequence
    b : list[T]
        b sequence

    Returns
    -------
    Callable[[float], T]
        Interpolator function

    Notes
    -----
    Use :func:`d3.interpolate_array <interpolate_array>` instead of this function.
    """
    nb = len(b) if b else 0
    na = min(nb, len(a)) if a else 0
    x = [interpolate(a[i], b[i]) for i in range(na)]
    c = list(b)

    def local_interpolate(t):
        for i in range(na):
            c[i] = x[i](t)
        return c

    return local_interpolate


def interpolate(a: T, b: T) -> Callable[[float], T]:
    """
    Returns an interpolator between the two arbitrary values a and b.

    .. list-table::
        :widths: 25 75

        *   - :code:`bool`
            - Returns a constant function based on b value
        *   - :code:`int` 
            - See :func:`d3.interpolate_number <interpolate_number>`
        *   - :code:`float`
            - See :func:`d3.interpolate_number <interpolate_number>`
        *   - :code:`str`
            - See :func:`d3.interpolate_rgb <interpolate_rgb>` if string can be \
              formatted as a color else for a generic string, see \
              :func:`d3.interpolate_string <interpolate_string>`
        *   - :class:`Color <detroit.color.color.Color>`
            - See :func:`d3.interpolate_rgb <interpolate_rgb>`
        *   - :code:`datetime`
            - See :func:`d3.interpolate_date <interpolate_date>`
        *   - :code:`list[int | float]`
            - See :func:`d3.interpolate_number_array <interpolate_number_array>`
        *   - :code:`list[T] | tuple[T]`
            - Returns an interpolator which recursively interpolates based on \
              :code:`T` values
        *   - :code:`dict[U, V]`
            - See :func:`d3.interpolate_object <interpolate_object>`
        *   - :code:`Any`
            - See :func:`d3.interpolate_number <interpolate_number>`

    Parameters
    ----------
    a : T
        Left bound of the interpolator
    b : T
        Right bound of the interpolator

    Returns
    -------
    Callable[[float], T]
        Interpolator function where input should be 0 and 1 and outputs a value
        between a and b:

        .. math::

            interpolator: [0, 1] & \\longrightarrow \\mathbb [a, b] \\\\
                               x & \\longmapsto y

    Examples
    --------

    >>> interpolator = d3.interpolate(10, 20)
    >>> interpolator(0)
    10.0
    >>> interpolator(1)
    20.0
    >>> interpolator(0.5)
    15.0
    """
    if b is None or isinstance(b, bool):
        return constant(b)
    if isinstance(b, (int, float)):
        return interpolate_number(a, b)
    if isinstance(b, str):
        c = color(b)
        if c:
            b = c
            return interpolate_rgb(a, b)
        else:
            return interpolate_string(a, b)
    if isinstance(b, Color):
        return interpolate_rgb(a, b)
    if isinstance(b, datetime):
        return interpolate_date(a, b)
    if is_number_array(b):
        return interpolate_number_array(a, b)
    if isinstance(b, (list, tuple)):
        return generic_array(a, b)
    if isinstance(b, dict):
        return interpolate_object(a, b)
    return interpolate_number(a, b)



from .arc import Arc
from .area import Area
from .area_radial import area_radial
from .curves import (
    curve_basis,
    curve_basis_closed,
    curve_basis_open,
    curve_bump_radial,
    curve_bump_x,
    curve_bump_y,
    curve_bundle,
    curve_cardinal,
    curve_cardinal_closed,
    curve_cardinal_open,
    curve_catmull_rom,
    curve_catmull_rom_closed,
    curve_catmull_rom_open,
    curve_linear,
    curve_linear_closed,
    curve_monotone_x,
    curve_monotone_y,
    curve_natural,
    curve_step,
    curve_step_after,
    curve_step_before,
)
from .line import Line
from .line_radial import line_radial
from .link import Link, link_horizontal, link_radial, link_vertical
from .offset import (
    offset_diverging,
    offset_expand,
    offset_none,
    offset_silhouette,
    offset_wiggle,
)
from .order import (
    order_appearance,
    order_ascending,
    order_descending,
    order_inside_out,
    order_none,
    order_reverse,
)
from .pie import Pie
from .point_radial import point_radial
from .stack import Stack
from .symbol import SYMBOLS_FILL, SYMBOLS_STROKE, Symbol
from .symbols import (
    symbol_asterisk,
    symbol_circle,
    symbol_cross,
    symbol_diamond,
    symbol_diamond2,
    symbol_plus,
    symbol_square,
    symbol_square2,
    symbol_star,
    symbol_times,
    symbol_triangle,
    symbol_triangle2,
    symbol_wye,
)

__all__ = [
    "Arc",
    "Area",
    "Line",
    "Link",
    "Pie",
    "SYMBOLS_FILL",
    "SYMBOLS_STROKE",
    "Stack",
    "Symbol",
    "area_radial",
    "curve_basis",
    "curve_basis_closed",
    "curve_basis_open",
    "curve_bump_radial",
    "curve_bump_x",
    "curve_bump_y",
    "curve_bundle",
    "curve_cardinal",
    "curve_cardinal_closed",
    "curve_cardinal_open",
    "curve_catmull_rom",
    "curve_catmull_rom_closed",
    "curve_catmull_rom_open",
    "curve_linear",
    "curve_linear_closed",
    "curve_monotone_x",
    "curve_monotone_y",
    "curve_natural",
    "curve_step",
    "curve_step_after",
    "curve_step_before",
    "line_radial",
    "link_horizontal",
    "link_radial",
    "link_vertical",
    "offset_diverging",
    "offset_expand",
    "offset_none",
    "offset_silhouette",
    "offset_wiggle",
    "order_appearance",
    "order_ascending",
    "order_descending",
    "order_inside_out",
    "order_none",
    "order_reverse",
    "point_radial",
    "symbol_asterisk",
    "symbol_circle",
    "symbol_cross",
    "symbol_diamond",
    "symbol_diamond2",
    "symbol_plus",
    "symbol_square",
    "symbol_square2",
    "symbol_star",
    "symbol_times",
    "symbol_triangle",
    "symbol_triangle2",
    "symbol_wye",
]



from ..types import Point2D
from .length import geo_length


def geo_distance(a: Point2D, b: Point2D) -> float:
    """
    Returns the great-arc distance in radians between the two points a and b.
    Each point must be specified as a two-element array :code:`[longitude,
    latitude]` in degrees. This is the spherical equivalent of
    :func:`GeoPath.measure <detroit.geo.path.path.GeoPath.measure>` given a
    LineString of two points.

    Parameters
    ----------
    a : Point2D
        2D point
    b : Point2D
        2D point

    Returns
    -------
    float
        Distance
    """
    return geo_length({"type": "LineString", "coordinates": [a, b]})



from collections.abc import Hashable, Iterable
from functools import reduce
from operator import iand, ior


def union(*iterables: Iterable[Hashable]) -> set[Hashable]:
    """
    Returns a new set containing every distinct value that appears in any of
    the givven iterables.

    Parameters
    ----------
    iterables : Iterable[Hashable]
        Iterables

    Returns
    -------
    set[Hashable]
        Set of all united values from iterables

    Examples
    --------
    >>> d3.union([1, 2], [3, 4], [5])
    {1, 2, 3, 4, 5}
    """
    return reduce(ior, map(set, iterables))


def difference(
    iterable: Iterable[Hashable], *others: Iterable[Hashable]
) -> set[Hashable]:
    """
    Returns a new set containing every value of iterable that is not in any of
    the other iterables.

    Parameters
    ----------
    iterable : Iterable[Hashable]
        Iterable to differenciate
    others : Iterable[Hashable]
        Other iterables

    Returns
    -------
    set[Hashable]
        Set of the iterable differenciated with other iterables

    Examples
    --------
    >>> d3.difference([1, 2, 3], [2, 5], [3])
    {1}
    """
    return set(iterable) - union(*others)


def intersection(*iterables: Iterable[Hashable]) -> set[Hashable]:
    """
    Returns a new set containing every distinct value that appears in all of
    the given iterables.

    Parameters
    ----------
    iterables : Iterable[Hashable]
        Iterables

    Returns
    -------
    set[Hashable]
        Set of all intersected iterables

    Examples
    --------
    >>> d3.intersection([1, 2, 3], [3, 4], [2, 3])
    {3}
    """
    return reduce(iand, map(set, iterables))


def superset(a: Iterable[Hashable], b: Iterable[Hashable]) -> bool:
    """
    Returns True if every value in the given iterable b is also in the given
    iterable a.

    Parameters
    ----------
    a : Iterable[Hashable]
        Iterable a
    b : Iterable[Hashable]
        Iterable b

    Returns
    -------
    bool
        True if a is a superset of b

    Examples
    --------
    >>> d3.superset([1, 2, 3], [1, 2])
    True
    >>> d3.superset([1, 2, 3], [4, 2])
    False
    """
    return not (bool(len(set(b) - set(a))))


def subset(a: Iterable[Hashable], b: Iterable[Hashable]) -> bool:
    """
    Returns True if every value in the given iterable a is also in the given
    iterable b.

    Parameters
    ----------
    a : Iterable[Hashable]
        Iterable a
    b : Iterable[Hashable]
        Iterable b

    Returns
    -------
    bool
        True if a is a subset of b

    Examples
    --------
    >>> d3.subset([1, 2], [1, 2, 3])
    True
    >>> d3.subset([4, 2], [1, 2, 3])
    False
    """
    return not (bool(len(set(a) - set(b))))


def disjoint(a: Iterable[Hashable], b: Iterable[Hashable]) -> bool:
    """
    Returns True if a and b contain no shared value

    Parameters
    ----------
    a : Iterable[Hashable]
        Iterable a
    b : Iterable[Hashable]
        Iterable b

    Returns
    -------
    bool
        True if a and b are disjoint

    Examples
    --------
    >>> d3.disjoint([1, 3], [2, 4])
    True
    >>> d3.disjoint([1, 3], [3, 4])
    False
    """
    return not (bool(len(set(a) & set(b))))



from __future__ import annotations

from collections.abc import Callable
from math import floor, inf, isfinite, isnan, nan

from ..array import argpass, extent, nice, ticks
from ..array.threshold import threshold_sturges
from ..types import MultiPolygonGeoJSON, Point2D
from .area import area
from .constant import constant
from .contains import contains

CASES = [
    [],
    [[[1.0, 1.5], [0.5, 1.0]]],
    [[[1.5, 1.0], [1.0, 1.5]]],
    [[[1.5, 1.0], [0.5, 1.0]]],
    [[[1.0, 0.5], [1.5, 1.0]]],
    [[[1.0, 1.5], [0.5, 1.0]], [[1.0, 0.5], [1.5, 1.0]]],
    [[[1.0, 0.5], [1.0, 1.5]]],
    [[[1.0, 0.5], [0.5, 1.0]]],
    [[[0.5, 1.0], [1.0, 0.5]]],
    [[[1.0, 1.5], [1.0, 0.5]]],
    [[[0.5, 1.0], [1.0, 0.5]], [[1.5, 1.0], [1.0, 1.5]]],
    [[[1.5, 1.0], [1.0, 0.5]]],
    [[[0.5, 1.0], [1.5, 1.0]]],
    [[[1.0, 1.5], [1.5, 1.0]]],
    [[[0.5, 1.0], [1.0, 1.5]]],
    [],
]


def sign(x: float) -> int:
    """
    Returns the sign of the input.

    Parameters
    ----------
    x : float
        Input value

    Returns
    -------
    int
        Sign value
    """
    return -1 if x < 0 else 1


def finite(x: float) -> float:
    """
    Returns :code:`x` if it is finite else :code:`nan`

    Parameters
    ----------
    x : float
        Input value

    Returns
    -------
    float
        Output value
    """
    return x if isfinite(x) else nan


def valid(v: float | None) -> float:
    """
    Returns :code:`v` if it is valid else :code:`-inf`

    Parameters
    ----------
    v : float | None
        Input value

    Returns
    -------
    float
        Output value
    """
    if v is None or isnan(v):
        return -inf
    else:
        return v


def above(x: float | None, value: float) -> bool:
    """
    Returns if :code:`x` is above :code:`value`

    Parameters
    ----------
    x : float | None
        Input value
    value : float
        Value to compare

    Returns
    -------
    bool
        Output value
    """
    return False if x is None else (x >= value)


def get(values: list[float], index: int) -> float | None:
    """
    Getter function for list of float. If the index does not exist, returns
    :code:`None`

    Parameters
    ----------
    values : list[float]
        List of value
    index : int
        Index

    Returns
    -------
    float | None
        Desired value given the index if it exists
    """
    index = int(index)
    return values[index] if index < len(values) else None


def smooth1(x: float, v0: float, v1: float, value: float) -> float:
    """
    Smooth function in one dimension

    Parameters
    ----------
    x : float
        Input value
    v0 : float
        Min bound value
    v1 : float
        Max bound value
    value : float
        Target value

    Returns
    -------
    float
        Smooth value
    """
    a = value - v0
    b = v1 - v0
    d = 0.0
    if isfinite(a) or isfinite(b):
        d = a / b if b != 0.0 else nan
    else:
        d = sign(a) / sign(b)
    return x if isnan(d) else x + d - 0.5


class Contours:
    """
    For each threshold value, the contour generator constructs a GeoJSON
    MultiPolygon geometry object representing the area where the input values
    are greater than or equal to the threshold value. The geometry is in planar
    coordinates, where :math:`(i + 0.5, j + 0.5)` corresponds to element
    :code:`i + j * n` in the input values array.
    """

    def __init__(self):
        self._dx = 1
        self._dy = 1
        self._threshold = threshold_sturges
        self._smooth = argpass(self._smooth_linear)

    def __call__(self, values: list[float]) -> list[MultiPolygonGeoJSON]:
        """
        Computes the contours for the given array of values, returning an array
        of GeoJSON MultiPolygon geometry objects.

        Parameters
        ----------
        values : list[float]
            Input grid values which must be an array of length :math:`n \\times
            m` where :math:`[n, m]` is the contour generator's size;
            furthermore, each :code:`values[i + j * n]` must represent the value
            at the position :math:`(i, j)`.

        Returns
        -------
        list[GeoJSON]
            Geometry objects
        """
        tz = self._threshold(values)
        if not isinstance(tz, list):
            e = extent(values, finite)
            tz = ticks(*nice(e[0], e[1], tz), tz)
            while tz[len(tz) - 1] >= e[1]:
                tz.pop()
            while tz[1] < e[0]:
                tz.pop(0)
        else:
            tz = sorted(tz)

        return list(map(lambda value: self.contour(values, value), tz))

    def contour(self, values: list[float], value: float) -> MultiPolygonGeoJSON:
        """
        Computes a single contour

        Parameters
        ----------
        values : list[float]
            Input grid values which must be an array of length :math:`n \\times
            m` where :math:`[n, m]` is the contour generator's size;
            furthermore, each :code:`values[i + j * n]` must represent the value
            at the position :math:`(i, j)`.
        value : float
            Threshold value; the input values are greater than or equal to this
            value.

        Returns
        -------
        GeoJSON
            GeoJSON MultiPolygon geometry object
        """
        v = nan if value is None else value
        if not isinstance(v, (int, float)) or isnan(v):
            raise ValueError(f"Invalid value: {value}")

        polygons = []
        holes = []

        def callback(ring: list[Point2D]):
            self._smooth(ring, values, v)
            if area(ring) > 0:
                polygons.append([ring])
            else:
                holes.append(ring)

        self._isorings(values, v, callback)
        for hole in holes:
            for polygon in polygons:
                if contains(polygon[0], hole) != -1:
                    polygon.append(hole)
                    break

        return {
            "type": "MultiPolygon",
            "value": value,
            "coordinates": polygons,
        }

    def _isorings(
        self,
        values: list[float],
        value: float,
        callback: Callable[[list[Point2D]], None],
    ):
        """
        Finds contour polygons and passes them to :code:`callback` function.

        Parameters
        ----------
        values : list[float]
            Input grid values
        value : float
            Threshold value
        callback : Callable[[list[Point2D]], None]
            Callback function called when a ring is found
        """
        fragment_by_start = {}
        fragment_by_end = {}

        x = y = -1

        def stitch(line: list[Point2D]):
            start = [line[0][0] + x, line[0][1] + y]
            end = [line[1][0] + x, line[1][1] + y]
            start_index = self._index(start)
            end_index = self._index(end)
            if f := fragment_by_end.get(start_index):
                if g := fragment_by_start.get(end_index):
                    fragment_by_end.pop(f["end"])
                    fragment_by_start.pop(g["start"])
                    if f == g:
                        f["ring"].append(end)
                        callback(f["ring"])
                    else:
                        fragment_by_start[f["start"]] = fragment_by_end[g["end"]] = {
                            "start": f["start"],
                            "end": g["end"],
                            "ring": f["ring"] + g["ring"],
                        }
                else:
                    fragment_by_end.pop(f["end"])
                    f["ring"].append(end)
                    f["end"] = end_index
                    fragment_by_end[end_index] = f
            elif f := fragment_by_start.get(end_index):
                if g := fragment_by_end.get(start_index):
                    fragment_by_start.pop(f["start"])
                    fragment_by_end.pop(g["end"])
                    if f == g:
                        f["ring"].append(end)
                        callback(f["ring"])
                    else:
                        fragment_by_start[g["start"]] = fragment_by_end[f["end"]] = {
                            "start": g["start"],
                            "end": f["end"],
                            "ring": g["ring"] + f["ring"],
                        }
                else:
                    fragment_by_start.pop(f["start"])
                    f["ring"].insert(0, start)
                    f["start"] = start_index
                    fragment_by_start[start_index] = f
            else:
                fragment_by_start[start_index] = fragment_by_end[end_index] = {
                    "start": start_index,
                    "end": end_index,
                    "ring": [start, end],
                }

        t1 = above(get(values, 0), value)
        for line in CASES[t1 << 1]:
            stitch(line)

        while x + 1 < self._dx - 1:
            x += 1
            t0 = t1
            t1 = above(get(values, x + 1), value)
            for line in CASES[t0 | t1 << 1]:
                stitch(line)
        x += 1
        for line in CASES[t1 << 0]:
            stitch(line)

        while y + 1 < self._dy - 1:
            y += 1
            x = -1
            t1 = above(get(values, y * self._dx + self._dx), value)
            t2 = above(get(values, y * self._dx), value)
            for line in CASES[t1 << 1 | t2 << 2]:
                stitch(line)
            while x + 1 < self._dx - 1:
                x += 1
                t0 = t1
                t1 = above(get(values, y * self._dx + self._dx + x + 1), value)
                t3 = t2
                t2 = above(get(values, y * self._dx + x + 1), value)
                for line in CASES[t0 | t1 << 1 | t2 << 2 | t3 << 3]:
                    stitch(line)
            x += 1
            for line in CASES[t1 | t2 << 3]:
                stitch(line)

        y += 1
        x = -1
        t2 = get(values, y * self._dx) >= value
        for line in CASES[t2 << 2]:
            stitch(line)
        while x + 1 < self._dx - 1:
            x += 1
            t3 = t2
            t2 = above(get(values, y * self._dx + x + 1), value)
            for line in CASES[t2 << 2 | t3 << 3]:
                stitch(line)

        x += 1
        for line in CASES[t2 << 3]:
            stitch(line)

    def _smooth_linear(self, ring: list[Point2D], values: list[float], value: float):
        """
        Function for smoothing contour polygons using linear interpolatation.

        Parameters
        ----------
        ring : list[Point2D]
            Contour polygons
        values : list[float]
            Input grid values
        value : float
            Threshold value
        """
        for point in ring:
            x = point[0]
            y = point[1]
            xt = int(x)
            yt = int(y)
            v1 = valid(get(values, yt * self._dx + xt))
            if x > 0 and x < self._dx and xt == x:
                point[0] = smooth1(
                    x, valid(get(values, yt * self._dx + xt - 1)), v1, value
                )
            if y > 0 and y < self._dy and yt == y:
                point[1] = smooth1(
                    y, valid(get(values, (yt - 1) * self._dx + xt)), v1, value
                )

    def _index(self, point: Point2D) -> int:
        """
        Returns the index given 2D point coordinates

        Parameters
        ----------
        point : Point2D
            2D point coordinates

        Returns
        -------
        int
            Index value
        """
        return int(point[0] * 2 + point[1] * (self._dx + 1) * 4)

    def set_size(self, size: tuple[float, float]) -> Contours:
        """
        Sets the size of the input values to the contour generator and returns
        itself.

        Parameters
        ----------
        size : tuple[float, float]
            Size of the input values

        Returns
        -------
        Contours
            Itself
        """
        dx = floor(size[0])
        dy = floor(size[1])
        if dx < 0.0 or dy < 0.0:
            raise ValueError("Invalid size: negative values")
        self._dx = dx
        self._dy = dy
        return self

    def set_thresholds(
        self,
        thresholds: Callable[[list[float], ...], float | list[float]]
        | list[float]
        | float,
    ) -> Contours:
        """
        Sets the threshold function to the contour generator and returns
        itself.

        Parameters
        ----------
        thresholds : Callable[[list[float], ...], float | list[float]] | list[float] | float
            Threshold function or array or constant value; default
            :func:`d3.threshold_sturges <detroit.threshold_sturges>`

        Returns
        -------
        Contours
            Itself
        """
        if callable(thresholds):
            self._threshold = thresholds
        else:
            self._threshold = constant(thresholds)
        return self

    def set_smooth(self, smooth: bool) -> Contours:
        """
        Sets if contour polygons are smoothed using interpolatation and returns
        itself.

        Parameters
        ----------
        smooth : bool
            Boolean value

        Returns
        -------
        Contours
            Itself
        """

        def noop():
            return

        self._smooth = argpass(self._smooth_linear if smooth else noop)
        return self

    def get_size(self) -> tuple[float, float]:
        return [self._dx, self._dy]

    def get_thresholds(self) -> Callable[[list[float], ...], float | list[float]]:
        return self._threshold

    def get_smooth(self) -> bool:
        return self._smooth == self._smooth_linear



def ease_cubic_in(t: float) -> float:
    """
    Cubic easing; equivalent to :code:`d3.ease_poly_in.exponent(3)`.

    Parameters
    ----------
    t : float
        :math:`t \\in [0, 1]`

    Returns
    -------
    float
        Output value of the easing function

    Examples
    --------
    >>> d3.ease_cubic_in(0.2)
    0.008000000000000002
    """
    return t * t * t


def ease_cubic_out(t: float) -> float:
    """
    Reverse cubic easing; equivalent to :code:`1 - d3.ease_cubic_in(1 - t)`.
    Also equivalent to :code:`d3.ease_poly_out.exponent(3)`.

    Parameters
    ----------
    t : float
        :math:`t \\in [0, 1]`

    Returns
    -------
    float
        Output value of the easing function

    Examples
    --------
    >>> d3.ease_cubic_out(0.2)
    0.4879999999999999
    """
    t -= 1
    return t * t * t + 1


def ease_cubic_in_out(t: float) -> float:
    """
    Symmetric cubic easing; scales :func:`d3.ease_cubic_in
    <detroit.ease_cubic_in>` :math:`\\forall t \\in [0, 0.5]` and
    :func:`d3.ease_cubic_out <detroit.ease_cubic_out>` :math:`\\forall t \\in
    [0.5, 1]`. Also equivalent to :code:`d3.ease_poly.exponent(3)`.

    Parameters
    ----------
    t : float
        :math:`t \\in [0, 1]`

    Returns
    -------
    float
        Output value of the easing function

    Examples
    --------
    >>> d3.ease_cubic_in_out(0.2)
    0.03200000000000001
    """
    t *= 2
    if t <= 1:
        return (t * t * t) * 0.5
    else:
        t -= 2
        return (t * t * t + 2) * 0.5



from __future__ import annotations

from bisect import bisect
from collections.abc import Callable
from datetime import datetime
from math import isnan, nan
from typing import Any, Generic

from ..interpolate import (
    interpolate as interpolate_value,
)
from ..interpolate import (
    interpolate_number,
    interpolate_round,
)
from ..types import GenValue, Number, T
from .utils import as_float, constant, identity


def normalize(a: GenValue, b: GenValue) -> Callable[[T], float]:
    """
    Makes a function which takes a generic value and returns the normalized
    value based on :code:`a` and :code:`b` values. Normalization is done by
    applying the following formula: :code:`(x - a) / (b - a)`.

    Parameters
    ----------
    a : GenValue
        Generic value
    b : GenValue
        Generic value

    Returns
    -------
    Callable[[T], float]
        Normalization function based on :code:`a` and :code:`b` values
    """
    a = as_float(a)
    b = as_float(b)
    b = b - a

    if isnan(b):
        return constant(nan)
    elif b == 0:
        return constant(0.5)

    def normalizer(x):
        x = as_float(x)
        return (x - a) / b

    return normalizer


def clamper(a: GenValue, b: GenValue) -> Callable[[GenValue], float]:
    """
    Makes a function which clamps its input value borned by :code:`a` and
    :code:`b` values.

    Parameters
    ----------
    a : GenValue
        Generic value
    b : GenValue
        Generic value

    Returns
    -------
    Callable[[GenValue], float]
        Clamp function based on :code:`a` and :code:`b` values
    """
    a = as_float(a)
    b = as_float(b)
    if a > b:
        a, b = b, a

    def clamp(x):
        x = as_float(x)
        return max(a, min(b, x))

    return clamp


class BiMap:
    def __init__(
        self,
        domain: tuple[T, T],
        range_vals: tuple[T, T],
        interpolate: Callable[[T, T], Callable[[T], float]],
    ):
        d0, d1 = domain[0], domain[1]
        r0, r1 = range_vals[0], range_vals[1]
        if d1 < d0:
            self.d0 = normalize(d1, d0)
            self.r0 = interpolate(r1, r0)
        else:
            self.d0 = normalize(d0, d1)
            self.r0 = interpolate(r0, r1)

    def __call__(self, x: T) -> float:
        return self.r0(self.d0(x))


class PolyMap:
    def __init__(
        self,
        domain: list[T],
        range_vals: list[T],
        interpolate: Callable[[T, T], Callable[[T], float]],
    ):
        self.domain = domain
        self.j = j = min(len(self.domain), len(range_vals)) - 1
        self.d = [None] * j
        self.r = [None] * j

        if self.domain[j] < self.domain[0]:
            self.domain = self.domain[::-1]
            range_vals = range_vals[::-1]

        for i in range(j):
            self.d[i] = normalize(self.domain[i], self.domain[i + 1])
            self.r[i] = interpolate(range_vals[i], range_vals[i + 1])

    def __call__(self, x: T) -> float:
        i = bisect(self.domain, x, 1, self.j) - 1
        return self.r[i](self.d[i](x))


def copy(source, target):
    return (
        target.set_domain(source.get_domain())
        .set_range(source.get_range())
        .set_interpolate(source.get_interpolate())
        .set_clamp(source.get_clamp())
        .set_unknown(source.get_unknown())
    )


class Transformer(Generic[T]):
    """
    Continous transformation

    Parameters
    ----------
    t : Callable[[Number], T]
        Transform function
    u : Callable[[T], Number]
        Untransform function
    """

    def __init__(
        self,
        t: Callable[[Number], T] = identity,
        u: Callable[[T], Number] = identity,
    ):
        self._transform = t
        self._untransform = u
        self._domain = [0, 1]
        self._range = [0, 1]
        self._clamp = identity
        self._interpolate = interpolate_value
        self._unknown = None
        self._input = None
        self._output = None
        self._rescale()

    def _rescale(self) -> Transformer:
        """
        Private method which updates:

        * :code:`clamp` attribute
        * :code:`piecewise` attribute
        * :code:`input` attribute
        * :code:`output` attribute
        """
        n = min(len(self._domain), len(self._range))
        if self._clamp != identity:
            self._clamp = clamper(self._domain[0], self._domain[n - 1])
        self.piecewise = PolyMap if n > 2 else BiMap
        self._output = self._input = None
        return self

    def __call__(self, x: Number) -> T:
        """
        Given a value from the domain, returns the corresponding value from the
        range.

        Parameters
        ----------
        x : Number
            Input value

        Returns
        -------
        T
            Corresponding value from the range
        """
        if x is None or (isinstance(x, float) and isnan(x)):
            return self._unknown
        else:
            if not self._output:
                domain = [
                    x.timestamp() if isinstance(x, datetime) else x
                    for x in self._domain
                ]
                domain = [self._transform(x) for x in domain]
                self._output = self.piecewise(domain, self._range, self._interpolate)
            return self._output(self._transform(self._clamp(x)))

    def invert(self, y: T) -> Number:
        """
        Given a value from the range, returns the corresponding value from the
        domain. Inversion is useful for interaction, say to determine the data
        value corresponding to the position of the mouse.

        Parameters
        ----------
        y : T
            Input value

        Returns
        -------
        Number
            Corresponding value from the domain
        """
        if not self._input:
            domain = [
                x.timestamp() if isinstance(x, datetime) else x for x in self._domain
            ]
            domain = [self._transform(x) for x in domain]
            self._input = self.piecewise(self._range, domain, interpolate_number)
        return self._clamp(self._untransform(self._input(y)))

    def set_domain(self, domain: list[Number]) -> Transformer:
        """
        Sets the scale's domain to the specified array of numbers.

        Parameters
        ----------
        domain : list[Number]
            Domain

        Returns
        -------
        Transformer
            Itself
        """
        # ODO: update lambda function to as_float function
        self._domain = list(
            map(lambda x: float(x) if isinstance(x, str) else x, domain)
        )
        return self._rescale()

    def get_domain(self) -> list[Number]:
        return self._domain.copy()

    def set_range(self, range_vals: list[T]) -> Transformer:
        """
        Sets the scale's range to the specified array of values

        Parameters
        ----------
        range_vals : list[T]
            Range values

        Returns
        -------
        Transformer
            Itself
        """
        self._range = list(range_vals)
        return self._rescale()

    def get_range(self) -> list[T]:
        return self._range.copy()

    def set_range_round(self, range_vals: list[T]) -> Transformer:
        """
        Sets the scale's range to the specified array of values and sets
        scale's interpolator to :code:`interpolate_round`.

        Parameters
        ----------
        range_vals : list[T]
            Range values

        Returns
        -------
        Transformer
            Itself
        """
        self._range = list(map(float, range_vals))
        self._interpolate = interpolate_round
        return self._rescale()

    def set_clamp(self, clamp: bool) -> Transformer:
        """
        Enables or disables clamping.

        Parameters
        ----------
        clamp : bool
            Clamp value

        Returns
        -------
        Transformer
            Itself
        """
        self._clamp = True if clamp else identity
        return self._rescale()

    def get_clamp(self) -> bool:
        return self._clamp != identity

    def set_interpolate(self, interpolate: Callable[[T, T], T]) -> Transformer:
        """
        Sets the scale's range interpolator factory.

        Parameters
        ----------
        interpolate : Callable[[T, T], T]
            Interpolate function

        Returns
        -------
        Transformer
            Itself
        """
        self._interpolate = interpolate
        return self._rescale()

    def get_interpolate(self) -> Callable[[T, T], T]:
        return self._interpolate

    def set_unknown(self, unknown: Any) -> Transformer:
        """
        Sets the output value of the scale for undefined or NaN input values.

        Parameters
        ----------
        unknown : Any
            Unknown value

        Returns
        -------
        Transformer
            Itself
        """
        self._unknown = unknown
        return self._rescale()

    def get_unknown(self) -> Any:
        return self._unknown

    def __str__(self) -> str:
        name = self.__class__.__name__
        attrbs = ["domain", "range"]
        attrbs = (f"{a}={getattr(self, f'get_{a}')()}" for a in attrbs)
        attrbs = ", ".join(attrbs)
        return f"{name}({attrbs})"

    def __repr__(self) -> str:
        name = self.__class__.__name__
        addr = id(self)
        return f"<{name} at {hex(addr)}>"
